# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Callable, Iterator
import inspect
from typing import Any, Concatenate, Generic, TypeVar

import grpc
from grpc import Channel
from packaging.version import parse as parse_version
from typing_extensions import ParamSpec, Self

from ansys.api.acp.v0.base_pb2 import CollectionPath, DeleteRequest, ListRequest

from ..._utils.property_protocols import ReadOnlyProperty
from ..._utils.resource_paths import join as _rp_join
from .._object_cache import ObjectCacheMixin, constructor_with_cache
from ..base import CreatableTreeObject, ServerWrapper, TreeObject, TreeObjectBase
from ..enums import Status
from .exceptions import wrap_grpc_errors
from .property_helper import _exposed_grpc_mapping_property, _wrap_doc
from .protocols import EditableAndReadableResourceStub, ObjectInfo, ReadableResourceStub

ValueT = TypeVar("ValueT", bound=TreeObjectBase)
CreatableValueT = TypeVar("CreatableValueT", bound=CreatableTreeObject)

__all__ = [
    "Mapping",
    "MutableMapping",
    "define_mutable_mapping",
    "define_create_method",
    "get_read_only_collection_property",
]


class Mapping(ObjectCacheMixin, Generic[ValueT]):
    """Mapping interface for collections of TreeObjects.

    Note: We could derive from collections.abc.Mapping to make sure
    this class conforms to the Mapping interface.
    """

    @classmethod
    @constructor_with_cache(
        key_getter=lambda *args, collection_path, **kwargs: collection_path.value,
        raise_on_invalid_key=True,
    )
    def _initialize_with_cache(
        cls,
        *,
        server_wrapper: ServerWrapper,
        collection_path: CollectionPath,
        stub: ReadableResourceStub,
        object_constructor: Callable[[ObjectInfo, Channel | None], ValueT],
    ) -> Self:
        return cls(
            _server_wrapper=server_wrapper,
            _collection_path=collection_path,
            _stub=stub,
            _object_constructor=object_constructor,
        )

    def __init__(
        self,
        *,
        _server_wrapper: ServerWrapper,
        _collection_path: CollectionPath,
        _stub: ReadableResourceStub,
        _object_constructor: Callable[[ObjectInfo, Channel | None], ValueT],
    ) -> None:
        self._collection_path = _collection_path
        self._stub = _stub

        self._server_wrapper = _server_wrapper
        self._object_constructor = _object_constructor

    @staticmethod
    def _cache_key_valid(key: Any) -> bool:
        if isinstance(key, str):
            return bool(key)
        return False

    def __iter__(self) -> Iterator[str]:
        yield from (obj.info.id for obj in self._get_objectinfo_list())

    def __getitem__(self, key: str) -> ValueT:
        obj_info = self._get_objectinfo_by_id(key)
        return self._object_constructor(obj_info, self._server_wrapper)

    def _get_objectinfo_list(self) -> list[ObjectInfo]:
        with wrap_grpc_errors():
            res = self._stub.List(ListRequest(collection_path=self._collection_path)).objects
        if len({obj.info.id for obj in res}) != len(res):
            raise ValueError("Duplicate ID in Collection.")
        return res

    def _get_objectinfo_by_id(self, key: str) -> ObjectInfo:
        for obj in self._get_objectinfo_list():
            if obj.info.id == key:
                return obj
        raise KeyError(f"No object with ID '{key}' found.")

    # def __setitem__(self, key: str, value: ValueT) -> None:
    #     raise NotImplementedError()

    # def update(self, other=(), /, **kwds):
    #     raise NotImplementedError()

    # def setdefault(self, key, default=None):
    #     raise NotImplementedError()

    def values(self) -> Iterator[ValueT]:
        """Return an iterator over the values of the mapping."""
        return (
            self._object_constructor(obj_info, self._server_wrapper)
            for obj_info in self._get_objectinfo_list()
        )

    def items(self) -> Iterator[tuple[str, ValueT]]:
        """Return an iterator over the (key, value) pairs of the mapping."""
        return (
            (
                obj_info.info.id,
                self._object_constructor(obj_info, self._server_wrapper),
            )
            for obj_info in self._get_objectinfo_list()
        )

    def keys(self) -> Iterator[str]:
        """Return an iterator over the keys of the mapping."""
        return iter(self)

    def __contains__(self, key: str) -> bool:
        """Return True if the mapping contains the given key."""
        return key in list(self)

    def __len__(self) -> int:
        """Return the number of items in the mapping."""
        return len(self._get_objectinfo_list())

    def get(self, key: str, default: ValueT | None = None) -> ValueT | None:
        """Return the value for key if key is in the mapping, else default."""
        try:
            return self[key]
        except KeyError:
            return default

    def __repr__(self) -> str:
        try:
            from ..object_registry import object_registry

            collection_label = self._collection_path.value.rsplit("/", 1)[-1]
            value_type = object_registry[collection_label]
            return f"<{self.__class__.__name__}[{value_type.__name__}] with keys {list(self)}>"
        except:
            return super().__repr__()


class MutableMapping(Mapping[CreatableValueT]):
    """Mutable mapping interface for collections of TreeObjects."""

    @classmethod
    @constructor_with_cache(
        key_getter=lambda *args, collection_path, **kwargs: collection_path.value,
        raise_on_invalid_key=True,
    )
    def _initialize_with_cache(
        cls,
        *,
        server_wrapper: ServerWrapper,
        collection_path: CollectionPath,
        stub: EditableAndReadableResourceStub,  # type: ignore # violates Liskov substitution
        object_constructor: Callable[[ObjectInfo, Channel | None], CreatableValueT],
    ) -> Self:
        return cls(
            _server_wrapper=server_wrapper,
            _collection_path=collection_path,
            _stub=stub,
            _object_constructor=object_constructor,
        )

    def __init__(
        self,
        *,
        _server_wrapper: ServerWrapper,
        _collection_path: CollectionPath,
        _stub: EditableAndReadableResourceStub,
        _object_constructor: Callable[[ObjectInfo, ServerWrapper | None], CreatableValueT],
    ) -> None:
        self._collection_path = _collection_path
        self._stub: EditableAndReadableResourceStub = _stub
        self._server_wrapper = _server_wrapper
        self._object_constructor = _object_constructor

    def __delitem__(self, key: str) -> None:
        obj_info = self._get_objectinfo_by_id(key)
        with wrap_grpc_errors():
            self._stub.Delete(
                DeleteRequest(
                    resource_path=obj_info.info.resource_path, version=obj_info.info.version
                )
            )

    def clear(self) -> None:
        """Remove all items from the mapping."""
        for obj_info in self._get_objectinfo_list():
            with wrap_grpc_errors():
                try:
                    self._stub.Delete(
                        DeleteRequest(
                            resource_path=obj_info.info.resource_path, version=obj_info.info.version
                        )
                    )
                except grpc.RpcError as e:
                    # The object version may have changed since the 'List' call,
                    # since the object may become outdated when other objects are deleted.
                    # In this case, we need to get the object info again and delete it.
                    if e.code() == grpc.StatusCode.FAILED_PRECONDITION:
                        obj_info = self._get_objectinfo_by_id(obj_info.info.id)
                        self._stub.Delete(
                            DeleteRequest(
                                resource_path=obj_info.info.resource_path,
                                version=obj_info.info.version,
                            )
                        )
                    else:
                        raise e

    def pop(self, key: str) -> CreatableValueT:
        """Remove and return the value for key."""
        obj_info = self._get_objectinfo_by_id(key)
        return self._pop_from_info(obj_info)

    def popitem(self) -> CreatableValueT:
        """Remove and return an arbitrary (key, value) pair from the mapping."""
        obj_info = self._get_objectinfo_list()[0]
        return self._pop_from_info(obj_info)

    def _pop_from_info(self, object_info: ObjectInfo) -> CreatableValueT:
        obj = self._object_constructor(object_info, self._server_wrapper)
        new_obj = obj.clone()
        obj.delete()
        return new_obj


ParentT = TypeVar("ParentT", bound=TreeObject)


def get_read_only_collection_property(
    object_class: type[ValueT],
    stub_class: type[ReadableResourceStub],
    doc: str | None = None,
    requires_uptodate: bool = False,
    supported_since: str | None = None,
) -> ReadOnlyProperty[Mapping[ValueT]]:
    """Define a read-only mapping of child tree objects."""

    def collection_property(self: ParentT) -> Mapping[ValueT]:
        if supported_since is not None and self._server_version is not None:
            if self._server_version < parse_version(supported_since):
                raise RuntimeError(
                    f"Accessing '{object_class.__name__}' objects on '{type(self).__name__}' "
                    f"requires version {supported_since} of the ACP gRPC server. The current server version is "
                    f"{self._server_version}."
                )
        if requires_uptodate and hasattr(self, "status") and not self.status == Status.UPTODATE:
            raise RuntimeError(
                f"The object {self.name} must be up-to-date to access {object_class.__name__}."
            )
        return Mapping._initialize_with_cache(
            server_wrapper=self._server_wrapper,
            collection_path=CollectionPath(
                value=_rp_join(self._resource_path.value, object_class._COLLECTION_LABEL)
            ),
            object_constructor=object_class._from_object_info,
            stub=stub_class(channel=self._server_wrapper.channel),
        )

    return _wrap_doc(
        _exposed_grpc_mapping_property(  # type: ignore
            collection_property, value_type=object_class, read_only=True
        ),
        doc=doc,
    )


P = ParamSpec("P")


def define_create_method(
    object_class: Callable[P, CreatableValueT],
    func_name: str,
    parent_class_name: str,
    module_name: str,
) -> Callable[Concatenate[ParentT, P], CreatableValueT]:
    """Define a create method for child tree objects."""

    def inner(self: ParentT, /, *args: P.args, **kwargs: P.kwargs) -> CreatableValueT:
        obj = object_class(*args, **kwargs)
        obj.store(parent=self)
        return obj

    # NOTE: This relies on our convention to document the tree object classes
    # on the class itself, instead of the __init__ method.
    inner.__doc__ = object_class.__doc__

    parameters = [inspect.signature(inner).parameters["self"]]
    parameters.extend(inspect.signature(object_class).parameters.values())
    inner.__signature__ = inspect.Signature(parameters, return_annotation=object_class)  # type: ignore

    inner.__name__ = func_name
    inner.__qualname__ = f"{parent_class_name}.{func_name}"
    inner.__module__ = module_name
    return inner


def define_mutable_mapping(
    object_class: type[CreatableValueT],
    stub_class: type[EditableAndReadableResourceStub],
    doc: str | None = None,
) -> ReadOnlyProperty[MutableMapping[CreatableValueT]]:
    """Define a mutable mapping of child tree objects."""

    def collection_property(self: ParentT) -> MutableMapping[CreatableValueT]:
        if self._server_version is not None:
            if self._server_version < parse_version(object_class._SUPPORTED_SINCE):
                raise RuntimeError(
                    f"The '{object_class.__name__}' object is only supported since version "
                    f"{object_class._SUPPORTED_SINCE} of the ACP gRPC server. The current server version is "
                    f"{self._server_version}."
                )

        return MutableMapping._initialize_with_cache(
            server_wrapper=self._server_wrapper,
            collection_path=CollectionPath(
                value=_rp_join(self._resource_path.value, object_class._COLLECTION_LABEL)
            ),
            object_constructor=object_class._from_object_info,
            stub=stub_class(channel=self._server_wrapper.channel),
        )

    return _wrap_doc(
        _exposed_grpc_mapping_property(
            collection_property, value_type=object_class, read_only=False  # type: ignore
        ),
        doc=doc,
    )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable, Sequence

from ansys.api.acp.v0 import (
    imported_modeling_ply_pb2,
    imported_modeling_ply_pb2_grpc,
    imported_production_ply_pb2_grpc,
)

from .._utils.property_protocols import ReadWriteProperty
from ._grpc_helpers.linked_object_list import define_linked_object_list
from ._grpc_helpers.mapping import get_read_only_collection_property
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    grpc_link_property,
    mark_grpc_properties,
)
from .base import CreatableTreeObject, IdTreeObject
from .enums import (
    ImportedPlyDrapingType,
    ImportedPlyOffsetType,
    ImportedPlyThicknessType,
    MeshImportType,
    RosetteSelectionMethod,
    ThicknessFieldType,
    imported_ply_draping_type_from_pb,
    imported_ply_draping_type_to_pb,
    imported_ply_offset_type_from_pb,
    imported_ply_offset_type_to_pb,
    imported_ply_thickness_type_from_pb,
    imported_ply_thickness_type_to_pb,
    mesh_import_type_from_pb,
    mesh_import_type_to_pb,
    rosette_selection_method_from_pb,
    rosette_selection_method_to_pb,
    status_type_from_pb,
    thickness_field_type_from_pb,
    thickness_field_type_to_pb,
)
from .fabric import Fabric
from .imported_production_ply import ImportedProductionPly
from .lookup_table_1d_column import LookUpTable1DColumn
from .lookup_table_3d_column import LookUpTable3DColumn
from .object_registry import register
from .rosette import Rosette
from .virtual_geometry import VirtualGeometry

__all__ = [
    "ImportedModelingPly",
]


@mark_grpc_properties
@register
class ImportedModelingPly(CreatableTreeObject, IdTreeObject):
    """Instantiate an Imported Modeling Ply.

    Parameters
    ----------
    name :
        The name of the Modeling Ply
    active :
        Inactive plies are ignored in ACP and the downstream analysis.
    offset_type :
        Defines the location of the mesh (plane) with respect to the laminate.
        One of ``BOTTOM_OFFSET``, ``MIDDLE_OFFSET``, and ``TOP_OFFSET``.
    mesh_import_type :
        Source of the imported mesh. Either from geometry or from hdf5.
    mesh_geometry:
        Link to the geometry with represents the ply surface. Only used if ``mesh_import_type``
        is ``FROM_GEOMETRY``.
    rosette_selection_method :
        Selection Method for Rosettes of the Oriented Selection Set.
    rosettes :
        Rosettes of the Oriented Selection Set.
    reference_direction_field :
        A 3D lookup table column that defines the reference directions.
    rotation_angle :
        Angle in degrees by which the initial reference directions are rotated around the orientations.
    ply_material :
        The material (only fabric is supported) of the ply.
    ply_angle :
        Design angle between the reference direction and the ply fiber direction.
    draping_type :
        Chooses between different draping formulations. ``NO_DRAPING`` by default.
    draping_angle_1_field :
        Correction angle between the fiber and draped fiber directions, in degree.
    draping_angle_2_field :
        Correction angle between the transverse and draped transverse directions,
        in degree. Optional, uses the same values as ``draping_angle_1_field``
        (no shear) by default.
    thickness_type :
        Choose :attr:`ThicknessType.FROM_TABLE` to define a ply with variable thickness.
        The default value is :attr:`ThicknessType.NOMINAL`, which means the ply
        thickness is constant and determined by the thickness of the ply material.
    thickness_field :
        Defines the look-up table column used to determine the ply thickness.
        Only applies if ``thickness_type`` is :attr:`ThicknessType.FROM_TABLE`.
    thickness_field_type :
        If ``thickness_type`` is :attr:`ThicknessType.FROM_TABLE`, this parameter
        determines how the thickness values are interpreted. They can be either
        absolute values (:attr:`ThicknessFieldType.ABSOLUTE_VALUES`) or relative
        values (:attr:`ThicknessFieldType.RELATIVE_SCALING_FACTOR`).
    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "imported_modeling_plies"
    _OBJECT_INFO_TYPE = imported_modeling_ply_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = imported_modeling_ply_pb2.CreateRequest
    _SUPPORTED_SINCE = "25.1"

    def __init__(
        self,
        *,
        name: str = "ModelingPly",
        active: bool = True,
        offset_type: ImportedPlyOffsetType = ImportedPlyOffsetType.MIDDLE_OFFSET,
        mesh_import_type: MeshImportType = MeshImportType.FROM_GEOMETRY,
        mesh_geometry: VirtualGeometry | None = None,
        rosette_selection_method: RosetteSelectionMethod = "minimum_angle",
        rosettes: Sequence[Rosette] = tuple(),
        reference_direction_field: LookUpTable3DColumn | LookUpTable1DColumn | None = None,
        rotation_angle: float = 0.0,
        ply_material: Fabric | None = None,
        ply_angle: float = 0.0,
        draping_type: ImportedPlyDrapingType = ImportedPlyDrapingType.NO_DRAPING,
        draping_angle_1_field: LookUpTable1DColumn | LookUpTable3DColumn | None = None,
        draping_angle_2_field: LookUpTable1DColumn | LookUpTable3DColumn | None = None,
        thickness_type: ImportedPlyThicknessType = ImportedPlyThicknessType.NOMINAL,
        thickness_field: LookUpTable1DColumn | LookUpTable3DColumn | None = None,
        thickness_field_type: ThicknessFieldType = ThicknessFieldType.ABSOLUTE_VALUES,
    ):
        super().__init__(name=name)

        self.active = active
        self.offset_type = ImportedPlyOffsetType(offset_type)
        self.mesh_import_type = MeshImportType(mesh_import_type)
        self.mesh_geometry = mesh_geometry
        self.rosette_selection_method = rosette_selection_method
        self.rosettes = rosettes
        self.reference_direction_field = reference_direction_field
        self.rotation_angle = rotation_angle

        self.ply_material = ply_material
        self.ply_angle = ply_angle

        self.draping_type = ImportedPlyDrapingType(draping_type)
        self.draping_angle_1_field = draping_angle_1_field
        self.draping_angle_2_field = draping_angle_2_field

        self.thickness_type = ImportedPlyThicknessType(thickness_type)
        self.thickness_field = thickness_field
        self.thickness_field_type = thickness_field_type

    def _create_stub(self) -> imported_modeling_ply_pb2_grpc.ObjectServiceStub:
        return imported_modeling_ply_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)
    active: ReadWriteProperty[bool, bool] = grpc_data_property("properties.active")
    offset_type = grpc_data_property(
        "properties.offset_type",
        from_protobuf=imported_ply_offset_type_from_pb,
        to_protobuf=imported_ply_offset_type_to_pb,
    )
    mesh_import_type = grpc_data_property(
        "properties.mesh_import_type",
        from_protobuf=mesh_import_type_from_pb,
        to_protobuf=mesh_import_type_to_pb,
    )
    mesh_geometry = grpc_link_property("properties.mesh_geometry", allowed_types=VirtualGeometry)
    rosette_selection_method = grpc_data_property(
        "properties.rosette_selection_method",
        from_protobuf=rosette_selection_method_from_pb,
        to_protobuf=rosette_selection_method_to_pb,
    )
    rosettes = define_linked_object_list("properties.rosettes", Rosette)
    reference_direction_field = grpc_link_property(
        "properties.reference_direction_field",
        allowed_types=(LookUpTable1DColumn, LookUpTable3DColumn),
    )
    rotation_angle: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.rotation_angle"
    )

    ply_material = grpc_link_property("properties.ply_material", allowed_types=(Fabric,))
    ply_angle: ReadWriteProperty[float, float] = grpc_data_property("properties.ply_angle")

    draping_type = grpc_data_property(
        "properties.draping",
        from_protobuf=imported_ply_draping_type_from_pb,
        to_protobuf=imported_ply_draping_type_to_pb,
    )
    draping_angle_1_field = grpc_link_property(
        "properties.draping_angle_1_field", allowed_types=(LookUpTable1DColumn, LookUpTable3DColumn)
    )
    draping_angle_2_field = grpc_link_property(
        "properties.draping_angle_2_field", allowed_types=(LookUpTable1DColumn, LookUpTable3DColumn)
    )

    thickness_type = grpc_data_property(
        "properties.thickness_type",
        from_protobuf=imported_ply_thickness_type_from_pb,
        to_protobuf=imported_ply_thickness_type_to_pb,
    )
    thickness_field = grpc_link_property(
        "properties.thickness_field", allowed_types=(LookUpTable1DColumn, LookUpTable3DColumn)
    )
    thickness_field_type = grpc_data_property(
        "properties.thickness_field_type",
        from_protobuf=thickness_field_type_from_pb,
        to_protobuf=thickness_field_type_to_pb,
    )

    imported_production_plies = get_read_only_collection_property(
        ImportedProductionPly, imported_production_ply_pb2_grpc.ObjectServiceStub
    )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Helper functions to show the structure of the ACP feature tree."""

from .._model_printer import Node
from .._tree_objects._grpc_helpers.property_helper import _exposed_grpc_mapping_property
from .._tree_objects.base import TreeObjectBase
from .._tree_objects.model import Model

__all__ = [
    "print_feature_tree",
    "get_feature_tree",
]


def print_feature_tree(show_lines: bool = False) -> None:
    """Print a tree representation of the PyACP features.

    Parameters
    ----------
    show_lines :
        Whether to show lines connecting the nodes.

    """
    print(get_feature_tree().to_string(show_lines=show_lines))


def get_feature_tree() -> Node:
    """Get a tree representation of the PyACP features."""
    return _get_feature_tree_impl(Model, False)


def _get_feature_tree_impl(root_cls: type[TreeObjectBase], read_only: bool) -> Node:
    label = root_cls.__name__
    if read_only:
        label += " (read-only)"
    grpc_properties = [getattr(root_cls, attr_name) for attr_name in root_cls._GRPC_PROPERTIES]
    return Node(
        label=label,
        children=[
            _get_feature_tree_impl(prop._value_type, read_only=prop._read_only)  # type: ignore
            for prop in grpc_properties
            if isinstance(prop, _exposed_grpc_mapping_property)
        ],
    )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing_extensions import Self

from ansys.api.acp.v0 import material_pb2

from ..._grpc_helpers.property_helper import mark_grpc_properties
from .base import (
    _ISOTROPIC_PROPERTY_UNAVAILABLE_MSG,
    _ORTHOTROPIC_PROPERTY_UNAVAILABLE_MSG,
    _ConstantPropertySet,
    _PolymorphicMixin,
    _PolymorphicPropertyKwargs,
    _VariablePropertySet,
)
from .property_helper import (
    constant_material_grpc_data_property,
    variable_material_grpc_data_property,
)

__all__ = [
    "ConstantEngineeringConstants",
    "VariableEngineeringConstants",
]


class _EngineeringConstantsMixin(_PolymorphicMixin):
    _DEFAULT_PB_PROPERTYSET_TYPE = material_pb2.OrthotropicEngineeringConstantsPropertySet
    _FIELD_NAME_DEFAULT = "_orthotropic"
    _FIELD_NAME_SUFFIX_BY_PB_DATATYPE = {
        material_pb2.IsotropicEngineeringConstantsPropertySet: "_isotropic",
        material_pb2.OrthotropicEngineeringConstantsPropertySet: "_orthotropic",
    }


_ISOTROPIC_KWARGS: _PolymorphicPropertyKwargs = {
    "available_on_pb_type": material_pb2.IsotropicEngineeringConstantsPropertySet,
    "unavailable_msg": _ISOTROPIC_PROPERTY_UNAVAILABLE_MSG,
}
_ORTHOTROPIC_KWARGS: _PolymorphicPropertyKwargs = {
    "available_on_pb_type": material_pb2.OrthotropicEngineeringConstantsPropertySet,
    "unavailable_msg": _ORTHOTROPIC_PROPERTY_UNAVAILABLE_MSG,
}


@mark_grpc_properties
class ConstantEngineeringConstants(_EngineeringConstantsMixin, _ConstantPropertySet):
    """Constant engineering constants material property set."""

    _GRPC_PROPERTIES = tuple()
    _SUPPORTED_SINCE = "24.2"

    @classmethod
    def from_isotropic_constants(
        cls,
        *,
        E: float = 0.0,
        nu: float = 0.0,  # correct??
    ) -> Self:
        """Create an isotropic engineering constants property set.

        Parameters
        ----------
        E :
            Young's modulus.
        nu :
            Poisson's ratio.

        Returns
        -------
        :
            An isotropic engineering constants property set.
        """
        obj = cls(
            _pb_object=cls._create_pb_object_from_propertyset_type(
                material_pb2.IsotropicEngineeringConstantsPropertySet
            )
        )
        obj.E = E
        obj.nu = nu
        return obj

    @classmethod
    def from_orthotropic_constants(
        cls,
        *,
        E1: float = 0.0,
        E2: float = 0.0,
        E3: float = 0.0,
        nu12: float = 0.0,
        nu23: float = 0.0,
        nu13: float = 0.0,
        G12: float = 0.0,
        G23: float = 0.0,
        G31: float = 0.0,
    ) -> Self:
        r"""Create an orthotropic engineering constants property set.

        Parameters
        ----------
        E1 :
            Young's modulus in material 1 direction.
        E2 :
            Young's modulus in material 2 direction.
        E3 :
            Young's modulus in out-of-plane direction.
        nu12 :
            Poisson's ratio :math:`\nu_{12}`.
        nu23 :
            Poisson's ratio :math:`\nu_{23}`.
        nu13 :
            Poisson's ratio :math:`\nu_{13}`.
        G12 :
            Shear modulus :math:`G_{12}`.
        G23 :
            Shear modulus :math:`G_{23}`.
        G31 :
            Shear modulus :math:`G_{31}`.

        Returns
        -------
        :
            An orthotropic engineering constants property set.
        """
        obj = cls()
        obj.E1 = E1
        obj.E2 = E2
        obj.E3 = E3
        obj.nu12 = nu12
        obj.nu23 = nu23
        obj.nu13 = nu13
        obj.G12 = G12
        obj.G23 = G23
        obj.G31 = G31
        return obj

    E = constant_material_grpc_data_property("E", **_ISOTROPIC_KWARGS)
    nu = constant_material_grpc_data_property("nu", **_ISOTROPIC_KWARGS)
    E1 = constant_material_grpc_data_property("E1", **_ORTHOTROPIC_KWARGS)
    E2 = constant_material_grpc_data_property("E2", **_ORTHOTROPIC_KWARGS)
    E3 = constant_material_grpc_data_property("E3", **_ORTHOTROPIC_KWARGS)
    G12 = constant_material_grpc_data_property("G12", **_ORTHOTROPIC_KWARGS)
    G23 = constant_material_grpc_data_property("G23", **_ORTHOTROPIC_KWARGS)
    G31 = constant_material_grpc_data_property("G31", **_ORTHOTROPIC_KWARGS)
    nu12 = constant_material_grpc_data_property("nu12", **_ORTHOTROPIC_KWARGS)
    nu23 = constant_material_grpc_data_property("nu23", **_ORTHOTROPIC_KWARGS)
    nu13 = constant_material_grpc_data_property("nu13", **_ORTHOTROPIC_KWARGS)


@mark_grpc_properties
class VariableEngineeringConstants(_EngineeringConstantsMixin, _VariablePropertySet):
    """Variable engineering constants material property set."""

    _GRPC_PROPERTIES = tuple()
    _SUPPORTED_SINCE = "24.2"

    E = variable_material_grpc_data_property("E", **_ISOTROPIC_KWARGS)
    nu = variable_material_grpc_data_property("nu", **_ISOTROPIC_KWARGS)
    E1 = variable_material_grpc_data_property("E1", **_ORTHOTROPIC_KWARGS)
    E2 = variable_material_grpc_data_property("E2", **_ORTHOTROPIC_KWARGS)
    E3 = variable_material_grpc_data_property("E3", **_ORTHOTROPIC_KWARGS)
    G12 = variable_material_grpc_data_property("G12", **_ORTHOTROPIC_KWARGS)
    G23 = variable_material_grpc_data_property("G23", **_ORTHOTROPIC_KWARGS)
    G31 = variable_material_grpc_data_property("G31", **_ORTHOTROPIC_KWARGS)
    nu12 = variable_material_grpc_data_property("nu12", **_ORTHOTROPIC_KWARGS)
    nu23 = variable_material_grpc_data_property("nu23", **_ORTHOTROPIC_KWARGS)
    nu13 = variable_material_grpc_data_property("nu13", **_ORTHOTROPIC_KWARGS)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Callable
import typing
from typing import TypeAlias, Union

from typing_extensions import Self

from ansys.api.acp.v0 import linked_selection_rule_pb2

from ._grpc_helpers.edge_property_list import GenericEdgePropertyType
from ._grpc_helpers.polymorphic_from_pb import tree_object_from_resource_path
from ._grpc_helpers.property_helper import _exposed_grpc_property, mark_grpc_properties
from .base import CreatableTreeObject
from .cut_off_selection_rule import CutOffSelectionRule
from .cylindrical_selection_rule import CylindricalSelectionRule
from .enums import (
    BooleanOperationType,
    boolean_operation_type_from_pb,
    boolean_operation_type_to_pb,
)
from .geometrical_selection_rule import GeometricalSelectionRule
from .parallel_selection_rule import ParallelSelectionRule
from .spherical_selection_rule import SphericalSelectionRule
from .tube_selection_rule import TubeSelectionRule
from .variable_offset_selection_rule import VariableOffsetSelectionRule

if typing.TYPE_CHECKING:  # pragma: no cover
    # Since the 'LinkedSelectionRule' class is used by the boolean selection rule,
    # this would cause a circular import at run-time.
    from .boolean_selection_rule import BooleanSelectionRule

_LINKABLE_SELECTION_RULE_TYPES: TypeAlias = Union[
    "BooleanSelectionRule",
    CutOffSelectionRule,
    CylindricalSelectionRule,
    GeometricalSelectionRule,
    ParallelSelectionRule,
    SphericalSelectionRule,
    TubeSelectionRule,
    VariableOffsetSelectionRule,
]


@mark_grpc_properties
class LinkedSelectionRule(GenericEdgePropertyType):
    r"""Defines selection rules linked to a Boolean Selection Rule or Modeling Ply.

    Parameters
    ----------
    selection_rule :
        Link to an existing selection rule.
    operation_type :
        Determines how the selection rule is combined with other selection rules.
    template_rule :
        If ``True``, the selection rule is a template rule. This means the parameter
        values are taken from the ``parameter_1`` and ``parameter_2`` attributes,
        not from the linked selection rule.
    parameter_1 :
        First parameter value of the selection rule. Only applies if ``template_rule``
        is ``True``.
    parameter_2 :
        Second parameter value of the selection rule. Only applies if ``template_rule``
        is ``True``.


    The ``parameter_1`` and ``parameter_2`` attributes determine the values of the
    following parameters, depending on the type of selection rule:

    ====================================== ================================== ===================
    Rule Type                              ``parameter_1``                    ``parameter_2``
    ====================================== ================================== ===================
    :class:`.ParallelSelectionRule`        ``lower_limit``                    ``upper_limit``
    :class:`.CylindricalSelectionRule`     ``radius``                         \-
    :class:`.SphericalSelectionRule`       ``radius``                         \-
    :class:`.TubeSelectionRule`            ``outer_radius``                   ``inner_radius``
    :class:`.GeometricalSelectionRule`     ``in_plane_capture_tolerance``     \-
    :class:`.VariableOffsetSelectionRule`  \-                                 \-
    :class:`.BooleanSelectionRule`         \-                                 \-
    ====================================== ================================== ===================

    Note that :class:`.CutOffSelectionRule` and :class:`.BooleanSelectionRule` objects cannot be linked to
    a Boolean Selection Rule, only to a Modeling Ply..
    """

    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        selection_rule: _LINKABLE_SELECTION_RULE_TYPES,
        *,
        operation_type: BooleanOperationType = BooleanOperationType.INTERSECT,
        template_rule: bool = False,
        parameter_1: float = 0.0,
        parameter_2: float = 0.0,
    ):
        # The '_callback_apply_changes' needs to be set first, otherwise the
        # setter methods will not work. We go through the setters instead of
        # directly setting the attributes to ensure the setter validation is
        # performed.
        self._callback_apply_changes: Callable[[], None] | None = None
        self.selection_rule = selection_rule
        self.operation_type = operation_type
        self.template_rule = template_rule
        self.parameter_1 = parameter_1
        self.parameter_2 = parameter_2

    @_exposed_grpc_property
    def selection_rule(self) -> _LINKABLE_SELECTION_RULE_TYPES:
        """Link to an existing selection rule."""
        return self._selection_rule

    @selection_rule.setter
    def selection_rule(self, value: _LINKABLE_SELECTION_RULE_TYPES) -> None:
        self._selection_rule = value
        if self._callback_apply_changes is not None:
            self._callback_apply_changes()

    @_exposed_grpc_property
    def operation_type(self) -> BooleanOperationType:
        """Operation to combine the selection rule with other selection rules."""
        return self._operation_type

    @operation_type.setter
    def operation_type(self, value: BooleanOperationType) -> None:
        # The backend converts the operation automatically; this is confusing
        # in the scripting context where the associated warning may not be visible.
        if isinstance(self._selection_rule, CutOffSelectionRule):
            if value != BooleanOperationType.INTERSECT:
                raise ValueError(
                    "Cannot use a boolean operation other than 'INTERSECT' with a "
                    "CutOffSelectionRule."
                )

        self._operation_type = value
        if self._callback_apply_changes is not None:
            self._callback_apply_changes()

    @_exposed_grpc_property
    def template_rule(self) -> bool:
        """Whether the selection rule is a template rule."""
        return self._template_rule

    @template_rule.setter
    def template_rule(self, value: bool) -> None:
        self._template_rule = value
        if self._callback_apply_changes is not None:
            self._callback_apply_changes()

    @_exposed_grpc_property
    def parameter_1(self) -> float:
        """First template parameter of the selection rule."""
        return self._parameter_1

    @parameter_1.setter
    def parameter_1(self, value: float) -> None:
        self._parameter_1 = value
        if self._callback_apply_changes is not None:
            self._callback_apply_changes()

    @_exposed_grpc_property
    def parameter_2(self) -> float:
        """Second template parameter of the selection rule."""
        return self._parameter_2

    @parameter_2.setter
    def parameter_2(self, value: float) -> None:
        self._parameter_2 = value
        if self._callback_apply_changes is not None:
            self._callback_apply_changes()

    def _set_callback_apply_changes(self, callback_apply_changes: Callable[[], None]) -> None:
        self._callback_apply_changes = callback_apply_changes

    @classmethod
    def _from_pb_object(
        cls,
        parent_object: CreatableTreeObject,
        message: linked_selection_rule_pb2.LinkedSelectionRule,
        apply_changes: Callable[[], None],
    ) -> Self:
        from .boolean_selection_rule import BooleanSelectionRule

        # Cannot link to objects of the same type as the parent.
        allowed_types_list = [
            ParallelSelectionRule,
            CylindricalSelectionRule,
            SphericalSelectionRule,
            TubeSelectionRule,
            GeometricalSelectionRule,
            VariableOffsetSelectionRule,
        ]
        if not isinstance(parent_object, BooleanSelectionRule):
            allowed_types_list += [CutOffSelectionRule, BooleanSelectionRule]
        allowed_types = tuple(allowed_types_list)

        selection_rule = tree_object_from_resource_path(
            resource_path=message.resource_path, server_wrapper=parent_object._server_wrapper
        )
        if not isinstance(selection_rule, allowed_types):
            raise TypeError(
                f"Expected selection_rule to be of type {allowed_types}, "
                f"got {type(selection_rule)}."
            )
        selection_rule = typing.cast("_LINKABLE_SELECTION_RULE_TYPES", selection_rule)
        new_obj = cls(
            selection_rule=selection_rule,
            operation_type=boolean_operation_type_from_pb(message.operation_type),
            template_rule=message.template_rule,
            parameter_1=message.parameter_1,
            parameter_2=message.parameter_2,
        )
        new_obj._set_callback_apply_changes(apply_changes)
        return new_obj

    def _to_pb_object(self) -> linked_selection_rule_pb2.LinkedSelectionRule:
        return linked_selection_rule_pb2.LinkedSelectionRule(
            resource_path=self.selection_rule._resource_path,
            operation_type=typing.cast(
                typing.Any, boolean_operation_type_to_pb(self.operation_type)
            ),
            template_rule=self.template_rule,
            parameter_1=self.parameter_1,
            parameter_2=self.parameter_2,
        )

    def _check(self) -> bool:
        # Check for empty resource paths
        return bool(self.selection_rule._resource_path.value)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, type(self)):
            return (
                self.selection_rule._resource_path == other.selection_rule._resource_path
                and self.operation_type == other.operation_type
                and self.template_rule == other.template_rule
                and self.parameter_1 == other.parameter_1
                and self.parameter_2 == other.parameter_2
            )

        return False

    def __repr__(self) -> str:
        return (
            f"LinkedSelectionRule(selection_rule={self.selection_rule.__repr__()}, "
            f"operation_type={self.operation_type}, "
            f"template_rule={self.template_rule}, "
            f"parameter_1={self.parameter_1}, "
            f"parameter_2={self.parameter_2})"
        )

    def clone(self) -> LinkedSelectionRule:
        """Create a new unstored LinkedSelectionRule with the same properties."""
        return LinkedSelectionRule(
            selection_rule=self.selection_rule,
            operation_type=self.operation_type,
            template_rule=self.template_rule,
            parameter_1=self.parameter_1,
            parameter_2=self.parameter_2,
        )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Callable
import os
import pathlib
import shutil
import typing
from typing import Any, Generic, Protocol, TypeVar, cast

import grpc

from ansys.api.acp.v0 import control_pb2_grpc, model_pb2_grpc
from ansys.api.acp.v0.base_pb2 import CollectionPath, DeleteRequest, Empty, ListRequest
from ansys.tools.filetransfer import Client as FileTransferClient

from .._tree_objects._grpc_helpers.exceptions import wrap_grpc_errors
from .._utils.typing_helper import PATH as _PATH
from .common import ServerKey, ServerProtocol

if typing.TYPE_CHECKING:  # pragma: no cover
    from .._tree_objects import Model


__all__ = ["ACPInstance"]


class FileTransferStrategy(Protocol):
    def upload_file(self, local_path: _PATH) -> pathlib.PurePath: ...

    def download_file(self, remote_path: _PATH, local_path: _PATH) -> None: ...

    def to_export_path(self, path: _PATH, is_directory: bool = False) -> pathlib.Path: ...


class LocalFileTransferStrategy(FileTransferStrategy):
    def __init__(self, working_directory: _PATH) -> None:
        self._working_directory = pathlib.Path(working_directory)

    def upload_file(self, local_path: _PATH) -> pathlib.Path:
        if not pathlib.Path(local_path).exists():
            raise FileNotFoundError(f"No such file or directory: '{local_path}'")
        return self._get_remote_path(local_path)

    def download_file(self, remote_path: _PATH, local_path: _PATH) -> None:
        remote_path_aslocal = self._get_local_path(remote_path)
        local_filename = pathlib.Path(local_path)
        if local_filename.exists() and local_filename.samefile(remote_path_aslocal):
            return
        shutil.copyfile(remote_path_aslocal, local_path)

    def to_export_path(self, path: _PATH, is_directory: bool = False) -> pathlib.Path:
        return self._get_remote_path(path)

    def _get_remote_path(self, path: _PATH) -> pathlib.Path:
        """Get the path in the format understood by the server.

        Convert the path from the format understood by the Python client to
        the format understood by the server.
        """
        return self._get_path_impl(path, pathlib.Path().cwd(), self._working_directory)

    def _get_local_path(self, path: _PATH) -> pathlib.Path:
        """Get the path in the format understood by the Python client.

        Convert the path from the format understood by the server to
        the format understood by the Python client.
        """
        return self._get_path_impl(path, self._working_directory, pathlib.Path().cwd())

    @staticmethod
    def _get_path_impl(
        path: _PATH,
        initial_working_directory: pathlib.Path,
        target_working_directory: pathlib.Path,
    ) -> pathlib.Path:
        path = pathlib.Path(path)
        if path.is_absolute() or initial_working_directory == target_working_directory:
            return path
        path = (initial_working_directory / path).resolve()
        try:
            return path.relative_to(target_working_directory)
        except ValueError:
            # If the path cannot be made relative (e.g. since it is on a different drive),
            # return the absolute path.
            return path


class RemoteFileTransferStrategy(FileTransferStrategy):
    def __init__(self, channel_getter: Callable[[], grpc.Channel]) -> None:
        self._channel_getter = channel_getter

    @property
    def _ft_client(self) -> FileTransferClient:
        return FileTransferClient(self._channel_getter())

    def upload_file(self, local_path: _PATH) -> pathlib.PurePath:
        remote_path = os.path.basename(local_path)
        self._ft_client.upload_file(local_filename=str(local_path), remote_filename=remote_path)
        return pathlib.PurePosixPath(remote_path)

    def download_file(self, remote_path: _PATH, local_path: _PATH) -> None:
        self._ft_client.download_file(
            remote_filename=str(remote_path), local_filename=str(local_path)
        )

    def to_export_path(self, path: _PATH, is_directory: bool = False) -> pathlib.Path:
        # Export to the working directory of the server
        if is_directory:
            return pathlib.Path(".")
        return pathlib.Path(pathlib.Path(path).name)


class FileTransferHandler:
    def __init__(
        self, filetransfer_strategy: FileTransferStrategy, auto_transfer_files: bool
    ) -> None:
        self._filetransfer_strategy = filetransfer_strategy
        self._auto_transfer_files = auto_transfer_files

    def upload_file_if_autotransfer(self, local_path: _PATH) -> pathlib.PurePath:
        if self._auto_transfer_files:
            return self.upload_file(local_path)
        return pathlib.Path(local_path)

    def download_file_if_autotransfer(self, remote_path: _PATH, local_path: _PATH) -> None:
        if self._auto_transfer_files:
            self.download_file(remote_path, local_path)
        else:
            # If auto-transfer is disabled, the export path should be the
            # same as the local path. Otherwise, this is a bug in our code.
            assert remote_path == local_path

    def to_export_path(self, path: _PATH, is_directory: bool = False) -> _PATH:
        if self._auto_transfer_files:
            return self._filetransfer_strategy.to_export_path(path, is_directory=is_directory)
        return path

    def upload_file(self, local_path: _PATH) -> pathlib.PurePath:
        return self._filetransfer_strategy.upload_file(local_path)

    def download_file(self, remote_path: _PATH, local_path: _PATH) -> None:
        self._filetransfer_strategy.download_file(remote_path, local_path)


ServerT = TypeVar("ServerT", bound=ServerProtocol, covariant=True)


class ACPInstance(Generic[ServerT]):
    """Control an ACP instance.

    Supports the following operations to control an ACP instance:

    - Loading and unloading ACP models
    - Up- and download of files to the (possibly remote) ACP instance
    - Check if the ACP instance is running

    In addition, the ACP instance *may* support starting, stopping,
    and restarting.

    Note that this class is not meant for instantiating directly.
    The :func:`.launch_acp` function should be used instead.
    """

    _server: ServerT
    _filetransfer_handler: FileTransferHandler
    _is_remote: bool

    def __init__(
        self,
        *,
        server: ServerT,
        filetransfer_handler: FileTransferHandler,
        is_remote: bool,
    ) -> None:
        self._server = server
        self._filetransfer_handler = filetransfer_handler
        self._is_remote = is_remote

    @property
    def _channel(self) -> grpc.Channel:
        return self._server.channels[ServerKey.MAIN]

    @property
    def is_remote(self) -> bool:
        """Whether the server is remote or local."""
        return self._is_remote

    @property
    def server_version(self) -> str:
        """Version of the connected server."""
        control_stub = control_pb2_grpc.ControlStub(self._channel)
        server_info = control_stub.GetServerInfo(Empty())
        version = server_info.version
        if not version:
            raise RuntimeError("Server version could not be determined.")
        return cast(str, version)

    def import_model(
        self,
        path: _PATH,
        *,
        name: str | None = None,
        format: str = "acp:h5",  # pylint: disable=redefined-builtin
        **kwargs: Any,
    ) -> Model:
        """Load an ACP model from a file.

        Parameters
        ----------
        name :
            Name of the newly loaded model.
        path :
            Path of the file to be loaded.
        format :
            Format of the file to be loaded. Can be one of ``"acp:h5"``,
            ``"ansys:h5"``, ``"ansys:cdb"``, ``"ansys:dat"``, ``"abaqus:inp"``,
            or ``"nastran:bdf"``.
        ignored_entities:
            Entities to ignore when loading the FE file. Can be a subset of
            the following values:
            ``"coordinate_systems"``, ``"element_sets"``, ``"materials"``,
            ``"mesh"``, or ``"shell_sections"``.
            Available only when the format is not ``"acp:h5"``.
        convert_section_data:
            Whether to import the section data of a shell model and convert it
            into ACP composite definitions.
            Available only when the format is not ``"acp:h5"``.
        unit_system:
            Defines the unit system of the imported file. Must be set if the
            input file does not have units. If the input file does have units,
            ``unit_system`` must be either ``"from_file"``, or match the input
            unit system.
            Available only when the format is not ``"acp:h5"``.

        Returns
        -------
        :
            The loaded ``Model`` instance.
        """
        from .._tree_objects import Model
        from .._tree_objects.base import ServerWrapper

        server_wrapper = ServerWrapper.from_acp_instance(self)
        if format == "acp:h5":
            if kwargs:
                raise ValueError(
                    f"Parameters '{kwargs.keys()}' cannot be passed when "
                    f"loading a model with format '{format}'."
                )
            model = Model._from_file(path=path, server_wrapper=server_wrapper)
        else:
            model = Model._from_fe_file(
                path=path,
                server_wrapper=server_wrapper,
                format=format,
                **kwargs,
            )
        if name is not None:
            model.name = name
        return model

    def clear(self) -> None:
        """Close all models.

        Closes the models which are currently open, without first
        saving them to a file.
        """
        from .._tree_objects import Model

        model_stub = model_pb2_grpc.ObjectServiceStub(self._channel)
        with wrap_grpc_errors():
            for model in model_stub.List(
                ListRequest(collection_path=CollectionPath(value=Model._COLLECTION_LABEL))
            ).objects:
                model_stub.Delete(
                    DeleteRequest(
                        resource_path=model.info.resource_path, version=model.info.version
                    )
                )

    @property
    def models(self) -> tuple[Model, ...]:
        """Models currently loaded on the server.

        Note that the models are listed in arbitrary order.
        """
        from .._tree_objects import Model

        model_stub = model_pb2_grpc.ObjectServiceStub(self._channel)
        return tuple(
            [
                Model._from_object_info(model_info, self._channel)
                for model_info in model_stub.List(
                    ListRequest(collection_path=CollectionPath(value=Model._COLLECTION_LABEL))
                ).objects
            ]
        )

    def upload_file(self, local_path: _PATH) -> pathlib.PurePath:
        """Upload a file to the server.

        .. warning::

            Do not execute this function with untrusted input parameters.
            See the :ref:`security guide<security_file_upload_download>`
            for details.

        Parameters
        ----------
        local_path :
            The path of the file to be uploaded.

        Returns
        -------
        :
            The path of the uploaded file on the server.
        """
        return self._filetransfer_handler.upload_file(local_path)

    def download_file(self, remote_path: _PATH, local_path: _PATH) -> None:
        """Download a file from the server.

        .. warning::

            Do not execute this function with untrusted input parameters.
            See the :ref:`security guide<security_file_upload_download>`
            for details.

        Parameters
        ----------
        remote_path :
            The path of the file on the server.
        local_path :
            The path of the file to be downloaded to.
        """
        self._filetransfer_handler.download_file(remote_path, local_path)

    def check(self, timeout: float | None = None) -> bool:
        """Check if the ACP instance is running.

        Parameters
        ----------
        timeout :
            Time to wait for the ACP instance to respond, in seconds. Note that
            there is no *guarantee* that ``check`` returns within this time.
            Instead, this parameter is used as a hint to the underlying
            implementation.
        """
        return self._server.check(timeout=timeout)

    def wait(self, timeout: float) -> None:
        """Wait for the ACP instance to respond.

        Repeatedly checks if the ACP instance is running, returning as
        soon as it is running.

        Parameters
        ----------
        timeout :
            Wait time before raising an exception.

        Raises
        ------
        RuntimeError
            In case the server still has not responded after ``timeout`` seconds.
        """
        self._server.wait(timeout=timeout)

    def start(self, timeout: float | None = None) -> None:
        """Start the product instance.

        Parameters
        ----------
        timeout :
            Timeout to wait until ACP responds. If ``None`` is specified,
            the check that ACP has started is skipped.

        Raises
        ------
        RuntimeError
            If the instance is already in the started state.
        RuntimeError
            If the instance does not allow manual starting.
        """
        if not hasattr(self._server, "start"):
            raise RuntimeError(
                "This ACP server does not expose a method to start it. "
                "Please use a different launch method."
            )
        self._server.start()
        if timeout is not None:
            self.wait(timeout=timeout)

    def stop(self, *, timeout: float | None = None) -> None:
        """Stop the product instance.

        Parameters
        ----------
        timeout :
            Time in seconds after which the instance is forcefully stopped. Note
            that not all launch methods may implement this parameter. If they
            do not, the parameter is ignored.

        Raises
        ------
        RuntimeError
            If the instance is already in the stopped state.
        RuntimeError
            If the instance does not allow stopping.
        """
        if not hasattr(self._server, "stop"):
            raise RuntimeError(
                "This ACP server does not expose a method to stop it. "
                "Please use a different launch method."
            )
        self._server.stop(timeout=timeout)

    def restart(
        self, stop_timeout: float | None = None, start_timeout: float | None = None
    ) -> None:
        """Stop, then start the product instance.

        Parameters
        ----------
        stop_timeout :
            Time in seconds after which the instance is forcefully stopped. Note
            that not all launch methods may implement this parameter. If they
            do not, the parameter is ignored.
        start_timeout :
            Timeout to wait until ACP responds. If ``None`` is specified,
            the check that ACP has started is skipped.

        Raises
        ------
        RuntimeError
            If the instance is already in the stopped state.
        RuntimeError
            If the instance does not allow restarting.
        """
        if not hasattr(self._server, "restart"):
            raise RuntimeError(
                "This ACP server does not expose a method to restart it. "
                "Please use a different launch method."
            )
        self._server.restart(stop_timeout=stop_timeout)
        if start_timeout is not None:
            self.wait(timeout=start_timeout)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import dataclasses

import grpc

from ansys.tools.local_product_launcher.helpers.grpc import check_grpc_health
from ansys.tools.local_product_launcher.interface import (
    METADATA_KEY_DOC,
    LauncherProtocol,
    ServerType,
)

from .common import ServerKey

__all__ = ["ConnectLaunchConfig"]


@dataclasses.dataclass
class ConnectLaunchConfig:
    """Configuration options for attaching to an existing ACP server."""

    url_acp: str = dataclasses.field(
        default="localhost:50555",
        metadata={METADATA_KEY_DOC: "URL to connect to for the main ACP server."},
    )
    url_filetransfer: str = dataclasses.field(
        default="localhost:50556",
        metadata={METADATA_KEY_DOC: "URL to connect to for the file transfer server."},
    )


class ConnectLauncher(LauncherProtocol[ConnectLaunchConfig]):
    CONFIG_MODEL = ConnectLaunchConfig
    SERVER_SPEC = {ServerKey.MAIN: ServerType.GRPC, ServerKey.FILE_TRANSFER: ServerType.GRPC}

    def __init__(self, *, config: ConnectLaunchConfig):
        self._config = config

    def start(self) -> None:
        # Since this launcher simply connects to an existing server, we don't need to start it.
        return

    def stop(self, *, timeout: float | None = None) -> None:
        # Since this launcher simply connects to an existing server, we don't need to stop it.
        return

    def check(self, timeout: float | None = None) -> bool:
        channel = grpc.insecure_channel(self.urls[ServerKey.MAIN])
        return check_grpc_health(channel=channel, timeout=timeout)

    @property
    def urls(self) -> dict[str, str]:
        return {
            ServerKey.MAIN: self._config.url_acp,
            ServerKey.FILE_TRANSFER: self._config.url_filetransfer,
        }



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:  # pragma: no cover
    import pyvista
    from ansys.acp.core import Model
    from ansys.acp.core.mesh_data import MeshData
    from ansys.acp.core.mesh_data import VectorData

from ._utils.pyvista_import_check import requires_pyvista
from ._utils.string_manipulation import replace_underscores_and_capitalize

__all__ = ["get_directions_plotter"]

_acp_direction_colors = {
    "normal": (1.0, 0.6, 0.0),
    "orientation": (1.0, 0, 1.0),
    "reference_direction": (1.0, 1.0, 0),
    "fiber_direction": (0.0, 1.0, 0.0),
    "transverse_direction": (0.0, 0.64, 0.11),
    "draped_fiber_direction": (0.0, 1.0, 1.0),
    "draped_transverse_direction": (0.0, 0.0, 1.0),
    "material_1_direction": (1.0, 0.0, 0.0),
}


@requires_pyvista
def get_directions_plotter(
    *,
    model: "Model",
    mesh: "MeshData | None" = None,
    components: Sequence[Optional["VectorData"]],
    culling_factor: int = 1,
    length_factor: float = 1.0,
    **kwargs: Any,
) -> "pyvista.Plotter":
    """Get a pyvista plotter that shows the specified directions on the mesh.

    Parameters
    ----------
    model :
        ACP model. Determines the average element size used for scaling the
        arrows. Unless explicitly specified, the model also determines the
        mesh to plot.
    mesh :
        Mesh defining the scope of the plot. If not provided, the full mesh
        of the model is used.
    components:
        List of components to plot.
    culling_factor :
        If set to a value other than ``1``, add only every n-th data
        point to the PyVista object. This is useful especially for
        vector data, where the arrows can be too dense.
    length_factor:
        Factor to scale the length of the arrows.
    kwargs :
        Keyword arguments passed to the PyVista object constructor.
    """
    import pyvista

    if mesh is None:
        mesh = model.mesh

    plotter: pyvista.Plotter = pyvista.Plotter()  # type: ignore
    plotter.add_mesh(mesh.to_pyvista(), color="white", show_edges=True)

    for vector_data in components:
        if vector_data is None:
            continue
        color = _acp_direction_colors.get(vector_data.component_name, "black")
        plotter.add_mesh(
            vector_data.get_pyvista_glyphs(
                mesh=mesh,
                factor=model.average_element_size * length_factor,
                culling_factor=culling_factor,
                **kwargs,
            ),
            color=color,
            label=replace_underscores_and_capitalize(vector_data.component_name),
        )
        plotter.add_legend(face=None, bcolor=[0.2, 0.2, 0.2], size=(0.25, 0.25))  # type: ignore
    return plotter



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from collections.abc import Iterator
from contextlib import contextmanager

from grpc import RpcError, StatusCode

STATUS_CODE_TO_EXCEPTION_TYPE = {
    StatusCode.INVALID_ARGUMENT: ValueError,
    StatusCode.NOT_FOUND: LookupError,
    StatusCode.ALREADY_EXISTS: RuntimeError,
    StatusCode.DEADLINE_EXCEEDED: TimeoutError,
    StatusCode.PERMISSION_DENIED: PermissionError,
    StatusCode.UNAUTHENTICATED: PermissionError,
    StatusCode.FAILED_PRECONDITION: RuntimeError,
    StatusCode.OUT_OF_RANGE: LookupError,
    StatusCode.UNIMPLEMENTED: NotImplementedError,
    StatusCode.UNAVAILABLE: ConnectionError,
    StatusCode.DATA_LOSS: RuntimeError,
    StatusCode.ABORTED: RuntimeError,
    StatusCode.INTERNAL: RuntimeError,
    StatusCode.UNKNOWN: RuntimeError,
}


@contextmanager
def wrap_grpc_errors() -> Iterator[None]:
    """Wrap gRPC errors in Python exceptions.

    This context manager will catch gRPC errors and re-raise them as Python
    exceptions with simpler error messages. The original error message will
    still be available in the traceback.
    """
    try:
        yield
    except RpcError as exc:
        details = exc.details().split("\n", 1)[0].strip()
        exception_type = STATUS_CODE_TO_EXCEPTION_TYPE.get(exc.code(), RuntimeError)
        raise exception_type(details) from exc



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import Any, TypedDict

from google.protobuf.message import Message

from ..._grpc_helpers.property_helper import grpc_data_property_read_only, mark_grpc_properties
from ...base import PolymorphicMixin as _BasePolymorphicMixin
from ...base import TreeObjectAttribute, TreeObjectAttributeReadOnly
from .variable_property_set_attributes import FieldVariable, InterpolationOptions

__all__ = (
    "_ConstantPropertySet",
    "_PolymorphicMixin",
    "_VariablePropertySet",
    "_ISOTROPIC_PROPERTY_UNAVAILABLE_MSG",
    "_ORTHOTROPIC_PROPERTY_UNAVAILABLE_MSG",
    "_PolymorphicPropertyKwargs",
)


_ISOTROPIC_PROPERTY_UNAVAILABLE_MSG = (
    "This property is only available for isotropic material property sets. "
    "The property set is currently orthotropic"
)
_ORTHOTROPIC_PROPERTY_UNAVAILABLE_MSG = (
    "This property is only available for orthotropic material property sets. "
    "The property set is currently isotropic"
)


class _PolymorphicPropertyKwargs(TypedDict):
    """Type for the extra keyword arguments for properties on polymorphic material property sets."""

    available_on_pb_type: type[Message]
    unavailable_msg: str


class _ConstantPropertySet(TreeObjectAttribute):
    """Base class for constant property sets."""

    _DEFAULT_PB_PROPERTYSET_TYPE: Any

    @classmethod
    def _create_default_pb_object(cls) -> Any:
        return cls._create_pb_object_from_propertyset_type(cls._DEFAULT_PB_PROPERTYSET_TYPE)

    @staticmethod
    def _create_pb_object_from_propertyset_type(pb_type: Any) -> Any:
        return pb_type(values=[pb_type.Data()])


@mark_grpc_properties
class _VariablePropertySet(TreeObjectAttributeReadOnly):
    """Base class for variable property sets."""

    field_variables = grpc_data_property_read_only(
        "field_variables",
        from_protobuf=lambda field_vars: tuple(
            FieldVariable(
                name=val.name,
                values=tuple(float(v) for v in val.values),
                default=val.default,
                lower_limit=val.lower_limit,
                upper_limit=val.upper_limit,
            )
            for val in field_vars
        ),
    )

    interpolation_options = grpc_data_property_read_only(
        "interpolation_options",
        from_protobuf=lambda val: InterpolationOptions(
            algorithm=val.algorithm, cached=val.cached, normalized=val.normalized
        ),
    )


class _PolymorphicMixin(_BasePolymorphicMixin):
    """Mixin class for property sets which can have multiple protobuf types.

    The class attributes defined here are consumed in the wrapper that exposes
    the property set on the parent Material.
    """

    _FIELD_NAME_DEFAULT: str
    _FIELD_NAME_SUFFIX_BY_PB_DATATYPE: dict[type[Message], str]



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable
import dataclasses

from ansys.api.acp.v0 import (
    variable_offset_selection_rule_pb2,
    variable_offset_selection_rule_pb2_grpc,
)

from .._utils.array_conversions import to_1D_double_array, to_tuple_from_1D_array
from .._utils.property_protocols import ReadWriteProperty
from ._elemental_or_nodal_data import (
    ElementalData,
    NodalData,
    VectorData,
    elemental_data_property,
    nodal_data_property,
)
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    grpc_link_property,
    mark_grpc_properties,
)
from ._mesh_data import full_mesh_property, shell_mesh_property
from .base import CreatableTreeObject, IdTreeObject
from .edge_set import EdgeSet
from .element_set import ElementSet
from .enums import status_type_from_pb
from .lookup_table_1d_column import LookUpTable1DColumn
from .object_registry import register

# Workaround: these imports are needed to make sphinx_autodoc_typehints understand
# the inherited members of the Elemental- and NodalData classes.
import numpy as np  # noqa: F401 isort:skip
from ._elemental_or_nodal_data import ScalarData  # noqa: F401 isort:skip

__all__ = [
    "VariableOffsetSelectionRule",
    "VariableOffsetSelectionRuleElementalData",
    "VariableOffsetSelectionRuleNodalData",
]


@dataclasses.dataclass
class VariableOffsetSelectionRuleElementalData(ElementalData):
    """Represents elemental data for a VariableOffset Selection Rule."""

    normal: VectorData | None = None


@dataclasses.dataclass
class VariableOffsetSelectionRuleNodalData(NodalData):
    """Represents nodal data for a VariableOffset Selection Rule."""


@mark_grpc_properties
@register
class VariableOffsetSelectionRule(CreatableTreeObject, IdTreeObject):
    """Instantiate a Variable Offset Selection Rule.

    Parameters
    ----------
    name :
        Name of the Variable Offset Selection Rule.
    edge_set :
        Defines the edge along which the rule will be applied to.
    offsets :
        Defines the in-plane offset. Cuts elements which are closer to the edge than this value.
    angles :
        Defines the angle between the reference surface and the cutting plane.
    include_rule :
        Include or exclude area in rule. Setting this to ``False``
        inverts the selection.
    use_offset_correction :
        If enabled, then the offset is measured along the surface of the selected Element Set.
    element_set :
        Defines the surface on which the offset correction is calculated on.
    inherit_from_lookup_table :
        Specifies whether to inherit the Look-Up Table object properties.
    radius_origin :
        Origin of the look-up table axis. Only applies if ``inherit_from_lookup_table``
        is ``False``.
    radius_direction :
        Direction of the look-up table axis. Only applies if ``inherit_from_lookup_table``
        is ``False``.
    distance_along_edge :
        If ``True``, the look-up locations are evaluated along the edge, instead of along
        the axis direction.
    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "variable_offset_selection_rules"
    _OBJECT_INFO_TYPE = variable_offset_selection_rule_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = variable_offset_selection_rule_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "VariableOffsetSelectionrule",
        edge_set: EdgeSet | None = None,
        offsets: LookUpTable1DColumn | None = None,
        angles: LookUpTable1DColumn | None = None,
        include_rule: bool = True,
        use_offset_correction: bool = False,
        element_set: ElementSet | None = None,
        inherit_from_lookup_table: bool = True,
        radius_origin: tuple[float, float, float] = (0.0, 0.0, 0.0),
        radius_direction: tuple[float, float, float] = (1.0, 0.0, 0.0),
        distance_along_edge: bool = False,
    ):
        super().__init__(name=name)
        self.edge_set = edge_set
        self.offsets = offsets
        self.angles = angles
        self.include_rule = include_rule
        self.use_offset_correction = use_offset_correction
        self.element_set = element_set
        self.inherit_from_lookup_table = inherit_from_lookup_table
        self.radius_origin = radius_origin
        self.radius_direction = radius_direction
        self.distance_along_edge = distance_along_edge

    def _create_stub(self) -> variable_offset_selection_rule_pb2_grpc.ObjectServiceStub:
        return variable_offset_selection_rule_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)

    edge_set = grpc_link_property("properties.edge_set", allowed_types=EdgeSet)
    offsets = grpc_link_property("properties.offsets", allowed_types=LookUpTable1DColumn)
    angles = grpc_link_property("properties.angles", allowed_types=LookUpTable1DColumn)
    include_rule: ReadWriteProperty[bool, bool] = grpc_data_property("properties.include_rule_type")
    use_offset_correction: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.use_offset_correction"
    )
    element_set = grpc_link_property("properties.element_set", allowed_types=ElementSet)
    inherit_from_lookup_table: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.inherit_from_lookup_table"
    )
    radius_origin = grpc_data_property(
        "properties.radius_origin",
        from_protobuf=to_tuple_from_1D_array,
        to_protobuf=to_1D_double_array,
    )
    radius_direction = grpc_data_property(
        "properties.radius_direction",
        from_protobuf=to_tuple_from_1D_array,
        to_protobuf=to_1D_double_array,
    )
    distance_along_edge: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.distance_along_edge"
    )

    mesh = full_mesh_property
    shell_mesh = shell_mesh_property
    # selection rules don't have solid mesh data
    elemental_data = elemental_data_property(VariableOffsetSelectionRuleElementalData)
    nodal_data = nodal_data_property(VariableOffsetSelectionRuleNodalData)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Helpers for defining material property sets.

Implements a helper function ``wrap_property_set`` to expose property sets on the
Material, as a property.
"""

from __future__ import annotations

from typing import Any, TypeVar

from ..._grpc_helpers.property_helper import _exposed_grpc_property
from ..._grpc_helpers.protocols import Editable
from .base import _ConstantPropertySet, _PolymorphicMixin, _VariablePropertySet

TC = TypeVar("TC", bound=_ConstantPropertySet)
TV = TypeVar("TV", bound=_VariablePropertySet)


__all__ = ["wrap_property_set"]


def _property_set_getter(name: str, type_constant: type[TC], type_variable: type[TV]) -> Any:
    assert issubclass(type_constant, _PolymorphicMixin) == issubclass(
        type_variable, _PolymorphicMixin
    )
    is_monomorphic = not issubclass(type_constant, _PolymorphicMixin)

    def inner(self: Editable) -> TC | TV | None:
        # self needs to be Editable because the init function
        # of type_constant and type_variable expect a Editable _parent_object
        self._get_if_stored()
        if is_monomorphic:
            if not self._pb_object.properties.property_sets.HasField(name):
                return None
            propset = getattr(self._pb_object.properties.property_sets, name)
        else:
            field_name = self._pb_object.properties.property_sets.WhichOneof(name)
            if field_name is None:
                return None
            propset = getattr(
                self._pb_object.properties.property_sets,
                field_name,
            )
        if (len(propset.values) > 1) or (len(propset.field_variables) > 0):
            return type_variable(
                _parent_object=self, _attribute_path=f"properties.property_sets.{name}"
            )
        if len(propset.values) == 0:
            return None
        return type_constant(
            _parent_object=self, _attribute_path=f"properties.property_sets.{name}"
        )

    return inner


def _property_set_setter(name: str, type_constant: type[TC], type_variable: type[TV]) -> Any:
    assert issubclass(type_constant, _PolymorphicMixin) == issubclass(
        type_variable, _PolymorphicMixin
    )
    is_monomorphic = not issubclass(type_constant, _PolymorphicMixin)

    def inner(self: Editable, value: type[TC] | None) -> None:
        self._get_if_stored()
        if isinstance(getattr(self, name), _VariablePropertySet):
            raise AttributeError("Cannot replace variable property sets.")
        if is_monomorphic:
            self._pb_object.properties.property_sets.ClearField(name)
        else:
            assert issubclass(type_constant, _PolymorphicMixin)
            for name_suffix in type_constant._FIELD_NAME_SUFFIX_BY_PB_DATATYPE.values():
                self._pb_object.properties.property_sets.ClearField(name + name_suffix)
        if value is not None:
            if is_monomorphic:
                pb_propset = getattr(self._pb_object.properties.property_sets, name)
            else:
                assert issubclass(type_constant, _PolymorphicMixin)
                name_suffix = type_constant._FIELD_NAME_SUFFIX_BY_PB_DATATYPE[
                    type(value._pb_object)
                ]
                pb_propset = getattr(self._pb_object.properties.property_sets, name + name_suffix)
            pb_propset.CopyFrom(value._pb_object)
        self._put_if_stored()

    return inner


def _property_set_deleter(name: str) -> Any:
    def inner(self: Any) -> None:
        setattr(self, name, None)

    return inner


def wrap_property_set(name: str, type_constant: type[TC], type_variable: type[TV]) -> Any:
    """Define a material property set on the material.

    Expose a property set on the ``Material``, as a property. The property
    automatically chooses the appropriate type (``type_constant`` or ``type_variable``),
    depending on whether the property set is constant (has exactly one value, and
    no field variables), or variable.

    For polymorphic property sets (which are defined as a protobuf 'oneof'), the
    wrapper takes care of storing / constructing the object with the currently
    active field.

    Parameters
    ----------
    name :
        Field name on which the property set is defined. If the property set is
        polymorphic (defined as a protobuf 'oneof'), this is the name of the
        'oneof' itself, not of the contained fields.
    type_constant :
        Type of the property set when the property set is constant (independent
        of field variables).
    type_variable :
        Type of the property set when the property set depends on field variables.
    """
    return (
        _exposed_grpc_property(
            _property_set_getter(
                name=name, type_constant=type_constant, type_variable=type_variable
            )
        )
        .setter(
            _property_set_setter(
                name=name, type_constant=type_constant, type_variable=type_variable
            )
        )
        .deleter(_property_set_deleter(name))
    )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Helper functions for converting ACP data to PyVista data structures."""

from enum import IntEnum

import numpy as np
import numpy.typing as npt
from pyvista.core.celltype import CellType as pv_CellType


class ElementType(IntEnum):
    """Enumeration of ACP element types."""

    VERTEX_ELEMENT = 100
    LINK2N3D = 101
    BEAM2N3D = 111

    SHELL3N = 121
    SHELL4N = 122
    LAYERED_SHELL4N = 123
    LAYERED_SHELL8N = 124
    LAYERED_SHELL3N = 125
    LAYERED_SHELL4N_QM6 = 126
    LAYERED_SHELL6N = 128
    LAYERED_SHELL4N_SIMPSON = 129

    TETRA4N = 131
    TETRA10N = 132
    PYRAMID5N = 133
    PYRAMID13N = 134

    HEXA8N = 141
    HEXA20N = 142
    PRISM6N = 143
    PRISM15N = 144

    PLANE6N_PLANE_STRAIN = 150
    PLANE6N_PLANE_STRESS = 151
    PLANE8N_PLANE_STRAIN = 152
    PLANE8N_PLANE_STRESS = 153

    LAYERED_HEXA8N = 161
    LAYERED_HEXA20N = 162
    LAYERED_PRISM6N = 163
    LAYERED_PRISM15N = 164

    DRAPING_SHELL4N = 170

    INTERFACE_ELEMENT6N = 180
    INTERFACE_ELEMENT8N = 181
    INTERFACE_ELEMENT12N = 182
    INTERFACE_ELEMENT16N = 183

    LAYERED_POLYHEDRON = 190


ELEMENT_TO_PYVISTA_TYPE = {
    ElementType.VERTEX_ELEMENT: pv_CellType.VERTEX,
    ElementType.LINK2N3D: pv_CellType.LINE,
    ElementType.BEAM2N3D: pv_CellType.LINE,
    ElementType.SHELL3N: pv_CellType.TRIANGLE,
    ElementType.SHELL4N: pv_CellType.QUAD,
    ElementType.LAYERED_SHELL4N: pv_CellType.QUAD,
    ElementType.LAYERED_SHELL8N: pv_CellType.QUADRATIC_QUAD,
    ElementType.LAYERED_SHELL3N: pv_CellType.TRIANGLE,
    ElementType.LAYERED_SHELL4N_QM6: pv_CellType.QUAD,
    ElementType.LAYERED_SHELL6N: pv_CellType.QUADRATIC_TRIANGLE,
    ElementType.LAYERED_SHELL4N_SIMPSON: pv_CellType.QUAD,
    ElementType.TETRA4N: pv_CellType.TETRA,
    ElementType.TETRA10N: pv_CellType.QUADRATIC_TETRA,
    ElementType.PYRAMID5N: pv_CellType.PYRAMID,
    ElementType.PYRAMID13N: pv_CellType.QUADRATIC_PYRAMID,
    ElementType.HEXA8N: pv_CellType.HEXAHEDRON,
    ElementType.HEXA20N: pv_CellType.QUADRATIC_HEXAHEDRON,
    ElementType.PRISM6N: pv_CellType.WEDGE,
    ElementType.PRISM15N: pv_CellType.QUADRATIC_WEDGE,
    ElementType.PLANE6N_PLANE_STRAIN: pv_CellType.QUADRATIC_TRIANGLE,
    ElementType.PLANE6N_PLANE_STRESS: pv_CellType.QUADRATIC_TRIANGLE,
    ElementType.PLANE8N_PLANE_STRAIN: pv_CellType.QUADRATIC_QUAD,
    ElementType.PLANE8N_PLANE_STRESS: pv_CellType.QUADRATIC_QUAD,
    ElementType.LAYERED_HEXA8N: pv_CellType.HEXAHEDRON,
    ElementType.LAYERED_HEXA20N: pv_CellType.QUADRATIC_HEXAHEDRON,
    ElementType.LAYERED_PRISM6N: pv_CellType.WEDGE,
    ElementType.LAYERED_PRISM15N: pv_CellType.QUADRATIC_WEDGE,
    ElementType.DRAPING_SHELL4N: pv_CellType.QUAD,
    ElementType.INTERFACE_ELEMENT6N: pv_CellType.WEDGE,
    ElementType.INTERFACE_ELEMENT8N: pv_CellType.HEXAHEDRON,
    ElementType.INTERFACE_ELEMENT12N: pv_CellType.TRIANGLE,
    ElementType.INTERFACE_ELEMENT16N: pv_CellType.QUAD,
    ElementType.LAYERED_POLYHEDRON: pv_CellType.POLYHEDRON,
}


def to_pyvista_faces(
    *,
    element_types: npt.NDArray[np.int32],
    element_nodes: npt.NDArray[np.int32],
    element_nodes_offsets: npt.NDArray[np.int32],
) -> npt.NDArray[np.int32]:
    """Convert ACP element data to PyVista faces."""
    target_size = element_types.size + element_nodes.size
    # InterfaceElement12N and InterfaceElement16N are not supported by VTK;
    # they are shown as shell 3n and 4n, respectively.
    INTERFACE_ELEMENT_12N = 182
    INTERFACE_ELEMENT_16N = 183
    target_size -= 9 * np.sum(element_types == INTERFACE_ELEMENT_12N)
    target_size -= 12 * np.sum(element_types == INTERFACE_ELEMENT_16N)

    faces = np.empty(target_size, dtype=np.int32)

    target_idx = 0
    for i, element_type in enumerate(element_types):
        start_idx = element_nodes_offsets[i]

        if element_type == INTERFACE_ELEMENT_12N:
            end_idx = start_idx + 3
        elif element_type == INTERFACE_ELEMENT_16N:
            end_idx = start_idx + 4
        else:
            try:
                end_idx = element_nodes_offsets[i + 1]
            except IndexError:
                end_idx = element_nodes.size

        current_size = end_idx - start_idx
        faces[target_idx] = current_size
        target_idx += 1
        faces[target_idx : target_idx + current_size] = element_nodes[start_idx:end_idx]
        target_idx += current_size
    return faces


def to_pyvista_types(element_types: npt.NDArray[np.int32]) -> npt.NDArray[np.int32]:
    """Convert ACP element types to PyVista cell types."""
    return np.array([ELEMENT_TO_PYVISTA_TYPE[el_type] for el_type in element_types])



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable
import dataclasses

from ansys.api.acp.v0 import geometrical_selection_rule_pb2, geometrical_selection_rule_pb2_grpc

from .._utils.property_protocols import ReadWriteProperty
from ._elemental_or_nodal_data import (
    ElementalData,
    NodalData,
    VectorData,
    elemental_data_property,
    nodal_data_property,
)
from ._grpc_helpers.linked_object_list import define_linked_object_list
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    grpc_link_property,
    mark_grpc_properties,
)
from ._mesh_data import full_mesh_property, shell_mesh_property
from .base import CreatableTreeObject, IdTreeObject
from .element_set import ElementSet
from .enums import (
    GeometricalRuleType,
    geometrical_rule_type_from_pb,
    geometrical_rule_type_to_pb,
    status_type_from_pb,
)
from .object_registry import register
from .virtual_geometry import VirtualGeometry

# Workaround: these imports are needed to make sphinx_autodoc_typehints understand
# the inherited members of the Elemental- and NodalData classes.
import numpy as np  # noqa: F401 isort:skip
from ._elemental_or_nodal_data import ScalarData  # noqa: F401 isort:skip

__all__ = [
    "GeometricalSelectionRule",
    "GeometricalSelectionRuleElementalData",
    "GeometricalSelectionRuleNodalData",
]


@dataclasses.dataclass
class GeometricalSelectionRuleElementalData(ElementalData):
    """Represents elemental data for a Geometrical Selection Rule."""

    normal: VectorData | None = None


@dataclasses.dataclass
class GeometricalSelectionRuleNodalData(NodalData):
    """Represents nodal data for a Geometrical Selection Rule."""


@mark_grpc_properties
@register
class GeometricalSelectionRule(CreatableTreeObject, IdTreeObject):
    """Instantiate a Geometrical Selection Rule.

    Parameters
    ----------
    name :
        Name of the Geometrical Selection Rule.
    geometrical_rule_type :
        Rule type. Can be either :attr:`.GeometricalRuleType.GEOMETRY` or
        :attr:`.GeometricalRuleType.ELEMENT_SETS`.
    geometry :
        Virtual geometry to use for the rule.
    element_sets :
        Element sets to use for the rule.
    include_rule :
        Include or exclude area in rule. Setting this to ``False``
        inverts the selection.
    use_default_tolerances :
        Whether to use default tolerances.
    in_plane_capture_tolerance :
        In-plane capture tolerance around the outline.
    negative_capture_tolerance :
        Capture tolerance along the surface or projection normal.
    positive_capture_tolerance :
        Capture tolerance along the opposite surface or projection normal.
    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "geometrical_selection_rules"
    _OBJECT_INFO_TYPE = geometrical_selection_rule_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = geometrical_selection_rule_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "GeometricalSelectionrule",
        geometrical_rule_type: GeometricalRuleType = GeometricalRuleType.GEOMETRY,
        geometry: VirtualGeometry | None = None,
        element_sets: Iterable[ElementSet] = (),
        include_rule: bool = True,
        use_default_tolerances: bool = True,
        in_plane_capture_tolerance: float = 0.0,
        negative_capture_tolerance: float = 0.0,
        positive_capture_tolerance: float = 0.0,
    ):
        super().__init__(name=name)
        self.geometrical_rule_type = geometrical_rule_type
        self.geometry = geometry
        self.element_sets = element_sets
        self.include_rule = include_rule
        self.use_default_tolerances = use_default_tolerances
        self.in_plane_capture_tolerance = in_plane_capture_tolerance
        self.negative_capture_tolerance = negative_capture_tolerance
        self.positive_capture_tolerance = positive_capture_tolerance

    def _create_stub(self) -> geometrical_selection_rule_pb2_grpc.ObjectServiceStub:
        return geometrical_selection_rule_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)

    geometrical_rule_type = grpc_data_property(
        "properties.geometrical_rule_type",
        from_protobuf=geometrical_rule_type_from_pb,
        to_protobuf=geometrical_rule_type_to_pb,
    )
    geometry = grpc_link_property("properties.geometry", allowed_types=VirtualGeometry)
    element_sets = define_linked_object_list("properties.element_sets", object_class=ElementSet)
    include_rule: ReadWriteProperty[bool, bool] = grpc_data_property("properties.include_rule_type")
    use_default_tolerances: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.use_default_tolerances"
    )
    in_plane_capture_tolerance: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.in_plane_capture_tolerance"
    )
    negative_capture_tolerance: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.negative_capture_tolerance"
    )
    positive_capture_tolerance: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.positive_capture_tolerance"
    )

    mesh = full_mesh_property
    shell_mesh = shell_mesh_property
    # selection rules don't have solid mesh data
    elemental_data = elemental_data_property(GeometricalSelectionRuleElementalData)
    nodal_data = nodal_data_property(GeometricalSelectionRuleNodalData)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import dataclasses
import typing

import numpy as np
import numpy.typing as npt
from packaging.version import parse as parse_version

if typing.TYPE_CHECKING:  # pragma: no cover
    from pyvista.core.pointset import UnstructuredGrid

from ansys.api.acp.v0 import base_pb2, mesh_query_pb2, mesh_query_pb2_grpc

from .._utils.array_conversions import to_numpy
from .._utils.property_protocols import ReadOnlyProperty
from .._utils.pyvista_import_check import requires_pyvista
from .base import TreeObject

__all__ = [
    "MeshData",
    "full_mesh_property",
    "shell_mesh_property",
    "solid_mesh_property",
]


@dataclasses.dataclass
class MeshData:
    """Container for the mesh data of an ACP Model."""

    node_labels: npt.NDArray[np.int32]
    node_coordinates: npt.NDArray[np.float64]
    element_labels: npt.NDArray[np.int32]
    element_types: npt.NDArray[np.int32]
    element_nodes: npt.NDArray[np.int32]
    element_nodes_offsets: npt.NDArray[np.int32]

    @requires_pyvista
    def to_pyvista(self) -> UnstructuredGrid:
        """Convert the mesh data to a PyVista mesh."""
        from pyvista.core.pointset import UnstructuredGrid

        from .._utils.visualization import to_pyvista_faces, to_pyvista_types

        return UnstructuredGrid(
            to_pyvista_faces(
                element_types=self.element_types,
                element_nodes=self.element_nodes,
                element_nodes_offsets=self.element_nodes_offsets,
            ),
            to_pyvista_types(self.element_types),
            self.node_coordinates,
        )


def _mesh_property_impl(
    element_scoping: mesh_query_pb2.ElementScopingType.ValueType, doc: str
) -> ReadOnlyProperty[MeshData]:
    def getter(self: TreeObject) -> MeshData:
        mesh_query_stub = mesh_query_pb2_grpc.MeshQueryServiceStub(self._channel)
        assert self._server_version is not None
        if self._server_version < parse_version("25.1"):
            from .model import Model

            if not isinstance(self, Model):
                raise RuntimeError(
                    "Mesh attributes for object types other than 'Model' are only supported "
                    "for server versions 25.1 and later."
                )
            if element_scoping != mesh_query_pb2.ElementScopingType.ALL:
                raise RuntimeError(
                    "Element scoping is only supported for server versions 25.1 and later."
                )
            request: base_pb2.GetRequest | mesh_query_pb2.GetMeshDataRequest = base_pb2.GetRequest(
                resource_path=self._resource_path
            )
        else:
            request = mesh_query_pb2.GetMeshDataRequest(
                resource_path=self._resource_path, element_scoping=element_scoping
            )
        reply = mesh_query_stub.GetMeshData(request)
        return MeshData(
            node_labels=to_numpy(reply.node_labels),
            node_coordinates=to_numpy(reply.node_coordinates),
            element_labels=to_numpy(reply.element_labels),
            element_types=to_numpy(reply.element_types),
            element_nodes=to_numpy(reply.element_nodes),
            element_nodes_offsets=to_numpy(reply.element_nodes_offsets),
        )

    return property(getter, doc=doc)


full_mesh_property = _mesh_property_impl(
    mesh_query_pb2.ElementScopingType.ALL, doc="Full mesh associated with the object."
)
shell_mesh_property = _mesh_property_impl(
    mesh_query_pb2.ElementScopingType.SHELL, doc="Shell mesh associated with the object."
)
solid_mesh_property = _mesh_property_impl(
    mesh_query_pb2.ElementScopingType.SOLID, doc="Solid mesh associated with the object."
)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from collections.abc import Callable, Iterable
from typing import Any, Concatenate, TypeAlias, TypeVar
from weakref import WeakValueDictionary

from typing_extensions import ParamSpec, Self

__all__ = ["ObjectCacheMixin", "constructor_with_cache"]

# Making the cache key type a TypeVar would require higher-kinded TypeVars,
# a feature not (yet) supported by the Python type system. See
#   https://github.com/python/typing/issues/548
# So for now, we alias it to 'Any'
_CACHE_KEY_T: TypeAlias = Any


class ObjectCacheMixin:
    """Mixin class to add an instance cache.

    Adds a cache to the class, enabling constructors decorated with
    ``constructor_with_cache`` to return an existing object instead of
    creating a new instance.
    """

    __slots__: Iterable[str] = ("__weakref__",)

    _OBJECT_CACHE: WeakValueDictionary[_CACHE_KEY_T, Self]

    def __init_subclass__(cls: type[Self]) -> None:
        cls._OBJECT_CACHE = WeakValueDictionary()
        super().__init_subclass__()

    @staticmethod
    def _cache_key_valid(key: _CACHE_KEY_T) -> bool:
        raise NotImplementedError("The _cache_key_valid implementation is missing")


T = TypeVar("T", bound=ObjectCacheMixin, covariant=True)
P = ParamSpec("P")

_WRAPPED_T: TypeAlias = Callable[Concatenate[type[T], P], T]


def constructor_with_cache(
    key_getter: Any, raise_on_invalid_key: bool = True
) -> Callable[[_WRAPPED_T[T, P]], _WRAPPED_T[T, P]]:
    """Enable using the object cache in a constructor classmethod.

    Parameters
    ----------
    key_getter :
        Returns the cache key, given the arguments to the constructor classmethod.
    raise_on_invalid_key :
        Whether an invalid key should be ignored, or raise an exception.
    """

    def decorator(func: _WRAPPED_T[T, P]) -> _WRAPPED_T[T, P]:
        def inner(cls: type[T], /, *args: P.args, **kwargs: P.kwargs) -> T:
            key = key_getter(*args, **kwargs)
            if cls._cache_key_valid(key):
                try:
                    return cls._OBJECT_CACHE[key]
                except KeyError:
                    pass
            else:
                if raise_on_invalid_key:
                    raise ValueError(f"Cache key '{key}' is invalid.")
            instance = func(cls, *args, **kwargs)
            cls._OBJECT_CACHE[key] = instance
            return instance

        return inner

    return decorator



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Extras of the Ansys Composites PrepPost module."""

from .example_helpers import (
    FLAT_PLATE_SHELL_CAMERA,
    FLAT_PLATE_SOLID_CAMERA,
    RACE_CARE_NOSE_CAMERA_METER,
    ExampleKeys,
    get_example_file,
    set_plot_theme,
)
from .feature_tree import get_feature_tree, print_feature_tree

__all__ = [
    "ExampleKeys",
    "FLAT_PLATE_SHELL_CAMERA",
    "FLAT_PLATE_SOLID_CAMERA",
    "get_example_file",
    "RACE_CARE_NOSE_CAMERA_METER",
    "set_plot_theme",
    "print_feature_tree",
    "get_feature_tree",
]



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Helper functions for exchanging data between PyACP and PyDPF or PyDPF - Composites."""

import typing

from ._tree_objects.enums import UnitSystemType

# Avoid dependencies on pydpf-composites and dpf-core if it is not used
if typing.TYPE_CHECKING:  # pragma: no cover
    from ansys.dpf.core import UnitSystem

__all__ = ["get_dpf_unit_system"]


def get_dpf_unit_system(unit_system: UnitSystemType) -> "UnitSystem":
    """Convert pyACP unit system to DPF unit system.

    Parameters
    ----------
    unit_system
        The pyACP unit system.
    """
    try:
        from ansys.dpf.core import unit_systems
    except ImportError as e:
        raise ImportError(
            "The pyACP unit system can only be converted to a DPF unit system if the "
            "ansys-dpf-core package is installed."
        ) from e

    unit_systems_map = {
        UnitSystemType.UNDEFINED: unit_systems.undefined,
        # looks like the only difference from MKS to SI is
        # that temperature is defined as Kelvin in SI and °C in MKS.
        # We should still force the user to use MKS in this case.
        UnitSystemType.SI: None,
        UnitSystemType.MKS: unit_systems.solver_mks,
        UnitSystemType.uMKS: unit_systems.solver_umks,
        UnitSystemType.CGS: unit_systems.solver_cgs,
        # MPA is equivalent to nmm
        UnitSystemType.MPA: unit_systems.solver_nmm,
        UnitSystemType.BFT: unit_systems.solver_bft,
        UnitSystemType.BIN: unit_systems.solver_bin,
    }

    if unit_systems_map[unit_system] is None:
        raise ValueError(f"Unit system {unit_system} not supported. Use MKS instead of SI.")
    if unit_system not in unit_systems_map:
        raise ValueError(f"Unit system {unit_system} not supported.")

    return unit_systems_map[unit_system]



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable
import dataclasses

from ansys.api.acp.v0 import cutoff_selection_rule_pb2, cutoff_selection_rule_pb2_grpc

from .._utils.property_protocols import ReadWriteProperty
from ._elemental_or_nodal_data import (
    ElementalData,
    NodalData,
    VectorData,
    elemental_data_property,
    nodal_data_property,
)
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    grpc_link_property,
    mark_grpc_properties,
)
from ._mesh_data import full_mesh_property, shell_mesh_property
from .base import CreatableTreeObject, IdTreeObject
from .edge_set import EdgeSet
from .enums import (
    CutOffRuleType,
    PlyCutOffType,
    cut_off_rule_type_from_pb,
    cut_off_rule_type_to_pb,
    ply_cut_off_type_from_pb,
    ply_cut_off_type_to_pb,
    status_type_from_pb,
)
from .object_registry import register
from .virtual_geometry import VirtualGeometry

# Workaround: these imports are needed to make sphinx_autodoc_typehints understand
# the inherited members of the Elemental- and NodalData classes.
import numpy as np  # noqa: F401 isort:skip
from ._elemental_or_nodal_data import ScalarData  # noqa: F401 isort:skip


__all__ = [
    "CutOffSelectionRule",
    "CutOffSelectionRuleElementalData",
    "CutOffSelectionRuleNodalData",
]


@dataclasses.dataclass
class CutOffSelectionRuleElementalData(ElementalData):
    """Represents elemental data for a Cut-Off Selection Rule."""

    normal: VectorData | None = None


@dataclasses.dataclass
class CutOffSelectionRuleNodalData(NodalData):
    """Represents nodal data for a Cut-Off Selection Rule."""


@mark_grpc_properties
@register
class CutOffSelectionRule(CreatableTreeObject, IdTreeObject):
    """Instantiate a Cut-Off Selection Rule.

    Parameters
    ----------
    name :
        Name of the Cut-Off Selection Rule.
    cut_off_rule_type :
        Determines if the cut-off is defined by a geometry or by a tapering edge.
    cut_off_geometry :
        Geometry used to define the cut-off. Only applies if
        ``cut_off_rule_type`` is GEOMETRY.
    taper_edge_set :
        Edge used to define the cut-off. Only applies if
        ``cut_off_rule_type`` is :attr:`.CutOffRuleType.TAPER`.
    offset :
        Moves the cutting plane along the out-of-plane direction. Always measured
        from the reference surface.
    angle :
        Defines the angle between the cutting plane and the reference surface.
    ply_cut_off_type :
        Either the complete production ply is cut-off
        (:attr:`PlyCutOffType.PRODUCTION_PLY_CUTOFF`) or individual analysis plies
        (:attr:`PlyCutOffType.ANALYSIS_PLY_CUTOFF`).
    ply_tapering :
        Whether the tapering of analysis plies is enabled.
    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "cutoff_selection_rules"
    _OBJECT_INFO_TYPE = cutoff_selection_rule_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = cutoff_selection_rule_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "CutOffSelectionrule",
        cut_off_rule_type: CutOffRuleType = CutOffRuleType.GEOMETRY,
        cut_off_geometry: VirtualGeometry | None = None,
        taper_edge_set: EdgeSet | None = None,
        offset: float = 0.0,
        angle: float = 0.0,
        ply_cut_off_type: PlyCutOffType = PlyCutOffType.PRODUCTION_PLY_CUTOFF,
        ply_tapering: bool = False,
    ):
        super().__init__(name=name)
        self.cut_off_rule_type = cut_off_rule_type
        self.cut_off_geometry = cut_off_geometry
        self.taper_edge_set = taper_edge_set
        self.offset = offset
        self.angle = angle
        self.ply_cut_off_type = ply_cut_off_type
        self.ply_tapering = ply_tapering

    def _create_stub(self) -> cutoff_selection_rule_pb2_grpc.ObjectServiceStub:
        return cutoff_selection_rule_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)

    cut_off_rule_type = grpc_data_property(
        "properties.cutoff_rule_type",
        from_protobuf=cut_off_rule_type_from_pb,
        to_protobuf=cut_off_rule_type_to_pb,
    )
    cut_off_geometry = grpc_link_property(
        "properties.cutoff_geometry", allowed_types=VirtualGeometry
    )
    taper_edge_set = grpc_link_property("properties.taper_edge_set", allowed_types=EdgeSet)
    offset: ReadWriteProperty[float, float] = grpc_data_property("properties.offset")
    angle: ReadWriteProperty[float, float] = grpc_data_property("properties.angle")
    ply_cut_off_type = grpc_data_property(
        "properties.ply_cutoff_type",
        from_protobuf=ply_cut_off_type_from_pb,
        to_protobuf=ply_cut_off_type_to_pb,
    )
    ply_tapering: ReadWriteProperty[bool, bool] = grpc_data_property("properties.ply_tapering")

    mesh = full_mesh_property
    shell_mesh = shell_mesh_property
    # selection rules don't have solid mesh data
    elemental_data = elemental_data_property(CutOffSelectionRuleElementalData)
    nodal_data = nodal_data_property(CutOffSelectionRuleNodalData)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from collections.abc import Callable
from functools import wraps
from typing import Concatenate, TypeAlias, TypeVar

from packaging.version import parse as parse_version
from typing_extensions import ParamSpec

from .protocols import Readable

T = TypeVar("T", bound=Readable)
P = ParamSpec("P")
R = TypeVar("R")
_WRAPPED_T: TypeAlias = Callable[Concatenate[T, P], R]


def supported_since(
    version: str | None, err_msg_tpl: str | None = None
) -> Callable[[_WRAPPED_T[T, P, R]], _WRAPPED_T[T, P, R]]:
    """Mark a TreeObjectBase method as supported since a specific server version.

    Raises an exception if the current server version does not match the required version.
    If either the given `version` or the server version is `None`, the decorator does nothing.

    Parameters
    ----------
    version :
        The server version since which the method is supported. If ``None``, the
        decorator does nothing.
    err_msg_tpl :
        A custom error message template. If ``None``, a default error message is used.
    """
    if version is None:
        # return a trivial decorator if no version is specified
        def trivial_decorator(func: _WRAPPED_T[T, P, R]) -> _WRAPPED_T[T, P, R]:
            return func

        return trivial_decorator

    required_version = parse_version(version)

    def decorator(func: _WRAPPED_T[T, P, R]) -> _WRAPPED_T[T, P, R]:
        @wraps(func)
        def inner(self: T, /, *args: P.args, **kwargs: P.kwargs) -> R:
            server_version = self._server_version
            # If the object is not stored, we cannot check the server version.
            if server_version is not None:
                if server_version < required_version:
                    if err_msg_tpl is None:
                        err_msg = (
                            f"The '{func.__name__}' method is only supported since version {version} "
                            f"of the ACP gRPC server. The current server version is {server_version}."
                        )
                    else:
                        err_msg = err_msg_tpl.format(
                            required_version=required_version, server_version=server_version
                        )
                    raise RuntimeError(err_msg)
            return func(self, *args, **kwargs)

        return inner

    return decorator



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import collections
from collections.abc import Iterator
import contextlib
import copy
import dataclasses
import importlib.resources
import math
import os
import pathlib
import subprocess  # nosec B404
import uuid

import grpc
from packaging.version import parse as parse_version

from ansys.tools.local_product_launcher.helpers.grpc import check_grpc_health
from ansys.tools.local_product_launcher.helpers.ports import find_free_ports
from ansys.tools.local_product_launcher.interface import (
    METADATA_KEY_DOC,
    METADATA_KEY_NOPROMPT,
    LauncherProtocol,
    ServerType,
)

from .common import ServerKey

__all__ = ["DockerComposeLaunchConfig"]


def _get_default_license_server() -> str:
    try:
        return os.environ["ANSYSLMD_LICENSE_FILE"]
    except KeyError:
        return ""


@dataclasses.dataclass
class DockerComposeLaunchConfig:
    """Configuration options for launching ACP through docker compose."""

    image_name_acp: str = dataclasses.field(
        default="ghcr.io/ansys/acp:latest",
        metadata={METADATA_KEY_DOC: "Docker image running the ACP gRPC server."},
    )
    image_name_filetransfer: str = dataclasses.field(
        default="ghcr.io/ansys/tools-filetransfer:latest",
        metadata={METADATA_KEY_DOC: "Docker image running the file transfer service."},
    )
    license_server: str = dataclasses.field(
        default=_get_default_license_server(),
        metadata={
            METADATA_KEY_DOC: (
                "License server passed to the container as "
                "'ANSYSLMD_LICENSE_FILE' environment variable."
            )
        },
    )
    keep_volume: bool = dataclasses.field(
        default=False,
        metadata={METADATA_KEY_DOC: "If true, keep the volume after docker compose is stopped."},
    )
    compose_file: str | None = dataclasses.field(
        default=None,
        metadata={
            METADATA_KEY_DOC: (
                "Docker compose file used to start the services. Uses the "
                "'docker-compose.yaml' shipped with PyACP by default."
            ),
            METADATA_KEY_NOPROMPT: True,
        },
    )
    environment_variables: dict[str, str] = dataclasses.field(
        default_factory=dict,
        metadata={
            METADATA_KEY_DOC: (
                "Additional environment variables passed to docker compose. These take "
                "precedence over environment variables defined through another configuration "
                "option (for example 'license_server' which defines 'ANSYSLMD_LICENSE_FILE') "
                "or the pre-existing environment variables."
            ),
            METADATA_KEY_NOPROMPT: True,
        },
    )


class DockerComposeLauncher(LauncherProtocol[DockerComposeLaunchConfig]):
    CONFIG_MODEL = DockerComposeLaunchConfig
    SERVER_SPEC = {ServerKey.MAIN: ServerType.GRPC, ServerKey.FILE_TRANSFER: ServerType.GRPC}

    def __init__(self, *, config: DockerComposeLaunchConfig):
        self._compose_name = f"pyacp_compose_{uuid.uuid4().hex}"
        self._urls: dict[str, str]

        try:
            import ansys.tools.filetransfer  # noqa
        except ImportError as err:
            raise ImportError(
                "The 'ansys.tools.filetransfer' module is needed to launch ACP via docker-compose."
            ) from err

        self._env = copy.deepcopy(os.environ)
        self._env.update(
            IMAGE_NAME_ACP=config.image_name_acp,
            IMAGE_NAME_FILETRANSFER=config.image_name_filetransfer,
            ANSYSLMD_LICENSE_FILE=config.license_server,
        )
        self._env.update(config.environment_variables)
        self._keep_volume = config.keep_volume

        if config.compose_file is not None:
            self._compose_file: pathlib.Path | None = pathlib.Path(config.compose_file)
        else:
            self._compose_file = None

        try:
            self._compose_version = parse_version(
                subprocess.check_output(  # nosec B603, B607: documented in 'security_considerations.rst'
                    ["docker", "compose", "version", "--short"], text=True
                ).replace(
                    "-", "+"
                )
            )
            self._compose_cmds = ["docker", "compose"]
        except subprocess.CalledProcessError:
            # If 'docker compose does not work, try 'docker-compose' instead.
            self._compose_version = parse_version(
                subprocess.check_output(  # nosec B603, B607: documented in 'security_considerations.rst'
                    ["docker-compose", "version", "--short"], text=True
                ).replace(
                    "-", "+"
                )
            )
            self._compose_cmds = ["docker-compose"]

    @contextlib.contextmanager
    def _get_compose_file(self) -> Iterator[pathlib.Path]:
        if self._compose_file is not None:
            yield self._compose_file
        else:
            with importlib.resources.path(__package__, "docker-compose.yaml") as compose_file:
                yield compose_file

    def start(self) -> None:
        with self._get_compose_file() as compose_file:
            port_acp, port_ft = find_free_ports(2)
            self._urls = {
                ServerKey.MAIN: f"localhost:{port_acp}",
                ServerKey.FILE_TRANSFER: f"localhost:{port_ft}",
            }

            env = collections.ChainMap(
                {"PORT_ACP": str(port_acp), "PORT_FILETRANSFER": str(port_ft)}, self._env
            )

            # The compose_file may be temporary, in particular if the package is a zipfile.
            # To avoid it being deleted before docker compose has read it, we use the '--wait'
            # flag for 'docker compose'.
            cmd = self._compose_cmds + [
                "-f",
                str(compose_file.resolve()),
                "--project-name",
                self._compose_name,
                "up",
                "--detach",
            ]
            if self._compose_version >= parse_version("2.1.1"):
                # The '--wait' flag is only available from version >= 2.1.1 of docker compose:
                # https://github.com/docker/compose/commit/72e4519cbfb6cdfc600e6ebfa377ce4b8e162c78
                cmd.append("--wait")
            subprocess.check_call(  # nosec B603: documented in 'security_considerations.rst'
                cmd, env=env, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
            )

    def stop(self, *, timeout: float | None = None) -> None:
        # The compose file needs to be passed for all commands with docker-compose 1.X.
        # With docker-compose 2.X, this no longer seems to be necessary.
        with self._get_compose_file() as compose_file:
            cmd = self._compose_cmds + [
                "-f",
                str(compose_file),
                "--project-name",
                self._compose_name,
                "down",
            ]
            if timeout is not None:
                # --timeout must be an integer, so we round up.
                cmd.extend(["--timeout", str(math.ceil(timeout))])
            if not self._keep_volume:
                cmd.append("--volumes")
            subprocess.check_call(  # nosec B603: documented in 'security_considerations.rst'
                cmd,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )

    def check(self, timeout: float | None = None) -> bool:
        for url in self.urls.values():
            channel = grpc.insecure_channel(url)
            if not check_grpc_health(channel=channel, timeout=timeout):
                return False
        return True

    @property
    def urls(self) -> dict[str, str]:
        return self._urls



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Move mesh_data into separate namespace.

Remove classes and function which are not used by the public API
into a separate namespace.
"""

from ._tree_objects import (
    AnalysisPlyElementalData,
    AnalysisPlyNodalData,
    BooleanSelectionRuleElementalData,
    BooleanSelectionRuleNodalData,
    CutOffSelectionRuleElementalData,
    CutOffSelectionRuleNodalData,
    CylindricalSelectionRuleElementalData,
    CylindricalSelectionRuleNodalData,
    ElementSetElementalData,
    ElementSetNodalData,
    GeometricalSelectionRuleElementalData,
    GeometricalSelectionRuleNodalData,
    ImportedSolidModelElementalData,
    ImportedSolidModelNodalData,
    MeshData,
    ModelElementalData,
    ModelingPlyElementalData,
    ModelingPlyNodalData,
    ModelNodalData,
    OrientedSelectionSetElementalData,
    OrientedSelectionSetNodalData,
    ParallelSelectionRuleElementalData,
    ParallelSelectionRuleNodalData,
    ProductionPlyElementalData,
    ProductionPlyNodalData,
    ScalarData,
    SolidElementSetElementalData,
    SolidElementSetNodalData,
    SolidModelElementalData,
    SolidModelNodalData,
    SphericalSelectionRuleElementalData,
    SphericalSelectionRuleNodalData,
    TriangleMesh,
    TubeSelectionRuleElementalData,
    TubeSelectionRuleNodalData,
    VariableOffsetSelectionRuleElementalData,
    VariableOffsetSelectionRuleNodalData,
    VectorData,
)

__all__ = [
    "AnalysisPlyElementalData",
    "AnalysisPlyNodalData",
    "BooleanSelectionRuleElementalData",
    "BooleanSelectionRuleNodalData",
    "CutOffSelectionRuleElementalData",
    "CutOffSelectionRuleNodalData",
    "CylindricalSelectionRuleElementalData",
    "CylindricalSelectionRuleNodalData",
    "ElementSetElementalData",
    "ElementSetNodalData",
    "GeometricalSelectionRuleElementalData",
    "GeometricalSelectionRuleNodalData",
    "ImportedSolidModelElementalData",
    "ImportedSolidModelNodalData",
    "MeshData",
    "ModelElementalData",
    "ModelingPlyElementalData",
    "ModelingPlyNodalData",
    "ModelNodalData",
    "OrientedSelectionSetElementalData",
    "OrientedSelectionSetNodalData",
    "ParallelSelectionRuleElementalData",
    "ParallelSelectionRuleNodalData",
    "ProductionPlyElementalData",
    "ProductionPlyNodalData",
    "ScalarData",
    "SolidElementSetElementalData",
    "SolidElementSetNodalData",
    "SolidModelElementalData",
    "SolidModelNodalData",
    "SphericalSelectionRuleElementalData",
    "SphericalSelectionRuleNodalData",
    "TriangleMesh",
    "TubeSelectionRuleElementalData",
    "TubeSelectionRuleNodalData",
    "VariableOffsetSelectionRuleElementalData",
    "VariableOffsetSelectionRuleNodalData",
    "VectorData",
]



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable

from ansys.api.acp.v0 import sampling_point_pb2, sampling_point_pb2_grpc

from .._utils.array_conversions import to_1D_double_array, to_tuple_from_1D_array
from .._utils.property_protocols import ReadOnlyProperty, ReadWriteProperty
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    grpc_link_property,
    mark_grpc_properties,
)
from .base import CreatableTreeObject, IdTreeObject
from .enums import status_type_from_pb
from .object_registry import register
from .rosette import Rosette

__all__ = ["SamplingPoint"]


@mark_grpc_properties
@register
class SamplingPoint(CreatableTreeObject, IdTreeObject):
    """Instantiate a Sampling Point.

    Parameters
    ----------
    name :
        Name of the sampling point.
    point :
        Coordinates of the sampling point.
    direction :
        Direction of the sampling point.
    use_default_reference_direction :
        Whether to use the element coordinate system when computing the laminate
        properties (CLT).
    rosette :
        Rosette defining the coordinate system used when computing the laminate
        properties (CLT). Only used when ``use_default_reference_direction`` is False.
    offset_is_middle :
        Activate this option to offset the reference surface to the mid-plane of the
        laminate for the computation of the ABD matrices and laminate properties (CLT).
    consider_coupling_effect :
        Whether the computation of the laminate engineering constants should consider
        the coupling effect if the B-Matrix of the ABD-Matrix is not zero. Computation
        of the ABD matrices is not affected by this parameter.

    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "sampling_points"
    _OBJECT_INFO_TYPE = sampling_point_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = sampling_point_pb2.CreateRequest
    _SUPPORTED_SINCE = "25.1"

    def __init__(
        self,
        *,
        name: str = "SamplingPoint",
        point: tuple[float, float, float] = (0.0, 0.0, 0.0),
        direction: tuple[float, float, float] = (0.0, 0.0, 0.0),
        use_default_reference_direction: bool = True,
        rosette: Rosette | None = None,
        offset_is_middle: bool = True,
        consider_coupling_effect: bool = True,
    ):
        super().__init__(name=name)

        self.point = point
        self.direction = direction
        self.use_default_reference_direction = use_default_reference_direction
        self.rosette = rosette
        self.offset_is_middle = offset_is_middle
        self.consider_coupling_effect = consider_coupling_effect

    def _create_stub(self) -> sampling_point_pb2_grpc.ObjectServiceStub:
        return sampling_point_pb2_grpc.ObjectServiceStub(self._channel)

    locked: ReadOnlyProperty[bool] = grpc_data_property_read_only("properties.locked")
    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)
    point = grpc_data_property(
        "properties.point", from_protobuf=to_tuple_from_1D_array, to_protobuf=to_1D_double_array
    )
    direction = grpc_data_property(
        "properties.direction", from_protobuf=to_tuple_from_1D_array, to_protobuf=to_1D_double_array
    )
    use_default_reference_direction: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.use_default_reference_direction"
    )
    rosette = grpc_link_property("properties.rosette", allowed_types=Rosette)
    reference_direction = grpc_data_property_read_only(
        "properties.reference_direction",
        from_protobuf=to_tuple_from_1D_array,
        doc="Local x-direction used in the computation of laminate properties (CLT).",
    )
    offset_is_middle: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.offset_is_middle"
    )
    consider_coupling_effect: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.consider_coupling_effect"
    )
